# horde.py
# Main interface for the horde to this library.
from __future__ import annotations

import base64
import io
import json
import random
import sys
import typing
from collections.abc import Callable
from copy import deepcopy
from enum import Enum, auto
from types import FunctionType

from horde_model_reference.meta_consts import STABLE_DIFFUSION_BASELINE_CATEGORY, get_baseline_native_resolution
from horde_sdk.ai_horde_api.apimodels import ImageGenerateJobPopResponse
from horde_sdk.ai_horde_api.apimodels.base import (
    GenMetadataEntry,
)
from horde_sdk.ai_horde_api.consts import KNOWN_FACEFIXERS, KNOWN_UPSCALERS, METADATA_TYPE, METADATA_VALUE
from loguru import logger
from PIL import Image
from pydantic import BaseModel

from hordelib.comfy_horde import Comfy_Horde
from hordelib.consts import MODEL_CATEGORY_NAMES
from hordelib.nodes.comfy_qr.qr_nodes import QRByModuleSizeSplitFunctionPatterns
from hordelib.shared_model_manager import SharedModelManager
from hordelib.utils.dynamicprompt import DynamicPromptParser
from hordelib.utils.image_utils import ImageUtils
from hordelib.utils.ioredirect import ComfyUIProgress


class ProgressState(Enum):
    """The state of the progress report"""

    started = auto()
    progress = auto()
    post_processing = auto()
    finished = auto()


class ProgressReport(BaseModel):
    """A progress message sent to a callback"""

    hordelib_progress_state: ProgressState
    comfyui_progress: ComfyUIProgress | None = None
    progress: float | None = None
    hordelib_message: str | None = None


class ResultingImageReturn:
    image: Image.Image | None
    rawpng: io.BytesIO | None
    faults: list[GenMetadataEntry]

    def __init__(
        self,
        image: Image.Image | None,
        rawpng: io.BytesIO | None,
        faults: list[GenMetadataEntry],
    ):
        if faults is None:
            faults = []

        for fault in faults:
            if not isinstance(fault, GenMetadataEntry):
                raise TypeError("faults must be a list of GenMetadataEntry")

        if image is not None and not isinstance(image, Image.Image):
            raise TypeError("image must be a PIL.Image.Image")

        if rawpng is not None and not isinstance(rawpng, io.BytesIO):
            raise TypeError("rawpng must be a io.BytesIO")

        self.image = image
        self.rawpng = rawpng
        self.faults = faults


def _calc_upscale_sampler_steps(
    payload: dict,
) -> int:
    """Use `ImageUtils.calc_upscale_sampler_steps(...)` to calculate the number of steps for the upscale sampler.

    Args:
        payload (dict): The payload to use for the calculation.

    Returns:
        int: The number of steps to use.
    """
    model_name = payload.get("model_name")
    baseline = None
    native_resolution = None
    if model_name is not None:
        baseline = SharedModelManager.model_reference_manager.stable_diffusion.get_model_baseline(model_name)
    if baseline is not None:
        try:
            baseline = STABLE_DIFFUSION_BASELINE_CATEGORY(baseline)
        except ValueError:
            baseline = None
            logger.warning(
                f"Model {model_name} has an invalid baseline {baseline} so we cannot calculate "
                "hires fix upscale steps.",
            )
        if baseline is not None:
            native_resolution = get_baseline_native_resolution(baseline)

    width: int | None = payload.get("width")
    height: int | None = payload.get("height")
    hires_fix_denoising_strength: float | None = payload.get("hires_fix_denoising_strength")
    ddim_steps: int | None = payload.get("ddim_steps")

    if width is None or height is None:
        raise ValueError("Width and height must be set to calculate upscale sampler steps")

    if hires_fix_denoising_strength is None:
        raise ValueError("Hires fix denoising strength must be set to calculate upscale sampler steps")

    if ddim_steps is None:
        raise ValueError("DDIM steps must be set to calculate upscale sampler steps")

    return ImageUtils.calc_upscale_sampler_steps(
        model_native_resolution=native_resolution,
        width=width,
        height=height,
        hires_fix_denoising_strength=hires_fix_denoising_strength,
        ddim_steps=ddim_steps,
    )


class HordeLib:
    _instance: HordeLib | None = None
    _initialised = False

    # Horde to comfy sampler mapping
    SAMPLERS_MAP = {
        "k_euler": "euler",
        "k_euler_a": "euler_ancestral",
        "k_heun": "heun",
        "k_dpm_2": "dpm_2",
        "k_dpm_2_a": "dpm_2_ancestral",
        "k_lms": "lms",
        "k_dpm_fast": "dpm_fast",
        "k_dpm_adaptive": "dpm_adaptive",
        "k_dpmpp_2s_a": "dpmpp_2s_ancestral",
        "k_dpmpp_sde": "dpmpp_sde",
        "k_dpmpp_2m": "dpmpp_2m",
        "ddim": "ddim",
        "uni_pc": "uni_pc",
        "uni_pc_bh2": "uni_pc_bh2",
        "plms": "euler",
        "lcm": "lcm",
    }

    # Horde names on the left, our node names on the right
    # We use this to dynamically route the image through the
    # right node by reconnect inputs.
    CONTROLNET_IMAGE_PREPROCESSOR_MAP = {
        "canny": "canny",
        "hed": "hed",
        "depth": "depth",
        "normal": "normal",
        "openpose": "openpose",
        "seg": "seg",
        "scribble": "scribble",
        "fakescribbles": "fakescribble",
        "hough": "mlsd",  # horde backward compatibility
        "mlsd": "mlsd",
        # "<unused>": "MiDaS-DepthMapPreprocessor",
        # "<unused>": "MediaPipe-HandPosePreprocessor",
        # "<unused>": "MediaPipe-FaceMeshPreprocessor",
        # "<unused>": "BinaryPreprocessor",
        # "<unused>": "ColorPreprocessor",
        # "<unused>": "PiDiNetPreprocessor",
    }

    CONTROLNET_MODEL_MAP = {
        "canny": "diff_control_sd15_canny_fp16.safetensors",
        "hed": "diff_control_sd15_hed_fp16.safetensors",
        "depth": "diff_control_sd15_depth_fp16.safetensors",
        "normal": "control_normal_fp16.safetensors",
        "openpose": "control_openpose_fp16.safetensors",
        "seg": "control_seg_fp16.safetensors",
        "scribble": "control_scribble_fp16.safetensors",
        "fakescribble": "control_scribble_fp16.safetensors",
        "mlsd": "control_mlsd_fp16.safetensors",
        "hough": "control_mlsd_fp16.safetensors",
    }

    SOURCE_IMAGE_PROCESSING_OPTIONS = ["img2img", "inpainting", "outpainting", "remix"]

    SCHEDULERS = ["normal", "karras", "simple", "ddim_uniform", "sgm_uniform", "exponential"]

    # Describe a valid payload, it's types and bounds. All incoming payload data is validated against,
    # and normalised to, this schema.
    PAYLOAD_SCHEMA = {
        "sampler_name": {"datatype": str, "values": list(SAMPLERS_MAP.keys()), "default": "k_euler"},
        "cfg_scale": {"datatype": float, "min": 1, "max": 100, "default": 8.0},
        "denoising_strength": {"datatype": float, "min": 0.01, "max": 1.0, "default": 1.0},
        "control_strength": {"datatype": float, "min": 0.01, "max": 3.0, "default": 1.0},
        "seed": {"datatype": int, "default": random.randint(0, sys.maxsize)},
        "width": {"datatype": int, "min": 64, "max": 8192, "default": 512, "divisible": 64},
        "height": {"datatype": int, "min": 64, "max": 8192, "default": 512, "divisible": 64},
        "hires_fix": {"datatype": bool, "default": False},
        "clip_skip": {"datatype": int, "min": 1, "max": 20, "default": 1},
        "control_type": {"datatype": str, "values": list(CONTROLNET_IMAGE_PREPROCESSOR_MAP.keys()), "default": None},
        "image_is_control": {"datatype": bool, "default": False},
        "return_control_map": {"datatype": bool, "default": False},
        "prompt": {"datatype": str, "default": ""},
        "negative_prompt": {"datatype": str, "default": ""},
        "loras": {"datatype": list, "default": []},
        "tis": {"datatype": list, "default": []},
        "ddim_steps": {"datatype": int, "min": 1, "max": 500, "default": 30},
        "n_iter": {"datatype": int, "min": 1, "max": 100, "default": 1},
        "model": {"datatype": str, "default": "stable_diffusion"},
        "source_mask": {"datatype": Image.Image, "default": None},
        "source_image": {"datatype": Image.Image, "default": None},
        "source_processing": {"datatype": str, "values": SOURCE_IMAGE_PROCESSING_OPTIONS, "default": None},
        "hires_fix_denoising_strength": {"datatype": float, "min": 0.01, "max": 1.0, "default": 0.65},
        "scheduler": {"datatype": str, "values": SCHEDULERS, "default": "normal"},
        "tiling": {"datatype": bool, "default": False},
        "model_name": {"datatype": str, "default": "stable_diffusion"},  # Used internally by hordelib
        "stable_cascade_stage_b": {"datatype": str, "default": None},  # Stable Cascade
        "stable_cascade_stage_c": {"datatype": str, "default": None},  # Stable Cascade
        "extra_source_images": {"datatype": list, "default": []},  # Stable Cascade Remix
        "extra_texts": {"datatype": list, "default": []},  # QR Codes (for now)
        "workflow": {"datatype": str, "default": "auto_detect"},
        "transparent": {"datatype": bool, "default": False},
    }

    EXTRA_IMAGES_SCHEMA = {
        "image": {"datatype": Image.Image, "default": None},
        "strength": {"datatype": float, "min": 0.0, "max": 5.0, "default": 1.0},
    }

    EXTRA_TEXTS_SCHEMA = {
        "text": {"datatype": str, "default": ""},
        "reference": {"datatype": str, "default": None},
    }

    LORA_SCHEMA = {
        "name": {"datatype": str, "default": ""},
        "model": {"datatype": float, "min": -10.0, "max": 10.0, "default": 1.0},
        "clip": {"datatype": float, "min": -10.0, "max": 10.0, "default": 1.0},
        "inject_trigger": {"datatype": str},
        "is_version": {"datatype": bool},
    }

    TIS_SCHEMA = {
        "name": {"datatype": str, "default": ""},
        "inject_ti": {"datatype": str},
        "strength": {"datatype": float, "min": -10, "max": 10.0, "default": 1.0},
    }

    # pipeline parameter <- hordelib payload parameter mapping
    PAYLOAD_TO_PIPELINE_PARAMETER_MAPPING: dict[str, str | Callable] = {  # FIXME
        "sampler.sampler_name": "sampler_name",
        "sampler.cfg": "cfg_scale",
        "sampler.denoise": "denoising_strength",
        "sampler.seed": "seed",
        "sampler.noise_seed": "seed",
        "empty_latent_image.height": "height",
        "empty_latent_image.width": "width",
        "sampler.scheduler": "scheduler",
        "clip_skip.stop_at_clip_layer": "clip_skip",
        "prompt.text": "prompt",
        "negative_prompt.text": "negative_prompt",
        "sampler.steps": "ddim_steps",
        "empty_latent_image.batch_size": "n_iter",
        "repeat_image_batch.amount": "n_iter",
        "model_loader.ckpt_name": "model",
        "model_loader.model_name": "model",
        "model_loader.horde_model_name": "model_name",
        "model_loader.seamless_tiling_enabled": "tiling",
        "image_loader.image": "source_image",
        "loras": "loras",
        "tis": "tis",
        "upscale_sampler.denoise": "hires_fix_denoising_strength",
        "upscale_sampler.seed": "seed",
        "upscale_sampler.cfg": "cfg_scale",
        "upscale_sampler.steps": _calc_upscale_sampler_steps,
        "upscale_sampler.sampler_name": "sampler_name",
        "controlnet_apply.strength": "control_strength",
        "controlnet_model_loader.control_net_name": "control_type",
        # Flux
        "cfg_guider.cfg": "cfg_scale",
        "random_noise.noise_seed": "seed",
        "k_sampler_select.sampler_name": "sampler_name",
        "basic_scheduler.denoise": "denoising_strength",
        "basic_scheduler.steps": "ddim_steps",
        # Stable Cascade
        "stable_cascade_empty_latent_image.width": "width",
        "stable_cascade_empty_latent_image.height": "height",
        "stable_cascade_empty_latent_image.batch_size": "n_iter",
        "sc_image_loader.image": "source_image",
        "sc_image_loader_0.image": "source_image",
        "sampler_stage_c.sampler_name": "sampler_name",
        "sampler_stage_b.sampler_name": "sampler_name",
        "sampler_stage_c.cfg": "cfg_scale",
        "sampler_stage_c.denoise": "denoising_strength",
        "sampler_stage_b.seed": "seed",
        "sampler_stage_c.seed": "seed",
        "sampler_stage_b.steps": "ddim_steps*0.33",
        "sampler_stage_c.steps": "ddim_steps*0.67",
        "model_loader_stage_c.ckpt_name": "stable_cascade_stage_c",
        "model_loader_stage_c.model_name": "stable_cascade_stage_c",
        "model_loader_stage_c.horde_model_name": "model_name",
        "model_loader_stage_b.ckpt_name": "stable_cascade_stage_b",
        "model_loader_stage_b.model_name": "stable_cascade_stage_b",
        "model_loader_stage_b.horde_model_name": "model_name",
        # Stable Cascade 2pass
        "2pass_sampler_stage_c.sampler_name": "sampler_name",
        "2pass_sampler_stage_c.steps": "ddim_steps*0.67",
        "2pass_sampler_stage_c.denoise": "hires_fix_denoising_strength",
        "2pass_sampler_stage_b.sampler_name": "sampler_name",
        "2pass_sampler_stage_b.steps": "ddim_steps*0.33",
        # QR Codes
        "sampler_bg.sampler_name": "sampler_name",
        "sampler_bg.cfg": "cfg_scale",
        "sampler_bg.denoise": "denoising_strength",
        "sampler_bg.seed": "seed",
        "sampler_bg.steps": "ddim_steps",
        "sampler_bg.noise_seed": "seed",
        "sampler_fg.sampler_name": "sampler_name",
        "sampler_fg.cfg": "cfg_scale",
        "sampler_fg.denoise": "denoising_strength",
        "sampler_fg.seed": "seed",
        "sampler_fg.steps": "ddim_steps",
        "sampler_fg.noise_seed": "seed",
        "controlnet_bg.strength": "control_strength",
        "solidmask_grey.width": "width",
        "solidmask_grey.height": "height",
        "solidmask_white.width": "width",
        "solidmask_white.height": "height",
        "solidmask_black.width": "width",
        "solidmask_black.height": "height",
        "qr_code_split.max_image_size": "width",
    }

    _comfyui_callback: Callable[[str, dict, str], None] | None = None

    # We are a singleton
    def __new__(
        cls,
        *,
        comfyui_callback: Callable[[str, dict, str], None] | None = None,
        aggressive_unloading: bool = True,
    ):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls._comfyui_callback = comfyui_callback
            cls.aggressive_unloading = aggressive_unloading

        return cls._instance

    # We initialise only ever once (in the lifetime of the singleton)
    def __init__(
        self,
        *,
        comfyui_callback: Callable[[str, dict, str], None] | None = None,
        aggressive_unloading: bool | None = True,
        # If you add any more parameters here, you should also add them to __new__ above
        # and follow the same pattern
    ):
        if not self._initialised:
            self.generator = Comfy_Horde(
                comfyui_callback=comfyui_callback if comfyui_callback else self._comfyui_callback,
                aggressive_unloading=(
                    aggressive_unloading if aggressive_unloading is not None else self.aggressive_unloading
                ),
            )
            self.__class__._initialised = True

    def _json_hack(self, obj):
        # Helper to serialise json which contains non-serialisable types
        if hasattr(obj, "__class__"):
            return f"{obj.__class__.__name__} instance"
        return f"Object of type {type(obj).__name__}"

    def dump_json(self, adict):
        logger.warning(json.dumps(adict, indent=4, default=self._json_hack))

    def _validate(self, value, datatype, min=None, max=None, default=None, values=None, divisible=None):
        """Check the given value against the given constraints. Return the fixed value."""

        # First, if we are passed no value at all, use the default
        if value is None:
            return default

        # We have a value, check the type
        if not isinstance(value, datatype):
            # try to coerce the type
            try:
                value = datatype(value)
            except (ValueError, TypeError):
                # epic fail, use the default
                return default

        # If the value must be divisible by some amount, assert that
        if divisible:
            if value % divisible != 0:
                value = ((value + (divisible - 1)) // divisible) * divisible

        # If we have a minimum, assert it
        if min is not None and value < min:
            value = min

        # If we have a maximum, assert it
        if max and value > max:
            value = max

        # If we have a list of allowed values, make sure the value is permitted
        if values:
            if isinstance(value, str):
                if value.lower() not in values:
                    # not an allowed value, use the default
                    return default
                value = value.lower()
            else:
                if value not in values:
                    # not an allowed value, use the default
                    return default

        return value

    def _validate_data_structure(self, data, schema_definition=PAYLOAD_SCHEMA):
        """Validate a data structure, assert parameters fall within the allowed bounds."""
        data = deepcopy(data)

        # Remove anything from the payload that isn't in our schema
        for key in data.copy().keys():
            if key.lower() not in schema_definition.keys():
                del data[key]

        # Build a valid schema payload, create attributes that don't exist using default values
        for key, schema in schema_definition.items():
            data[key.lower()] = self._validate(data.get(key.lower()), **schema)

        # Do the same for loras, if we have loras in this data structure
        if data.get("loras"):
            for i, lora in enumerate(data.get("loras")):
                data["loras"][i] = self._validate_data_structure(lora, HordeLib.LORA_SCHEMA)
            # Remove invalid loras
            data["loras"] = [x for x in data["loras"] if x.get("name")]

        # Do the same for tis, if we have tis in this data structure
        if data.get("tis"):
            for i, ti in enumerate(data.get("tis")):
                data["tis"][i] = self._validate_data_structure(ti, HordeLib.TIS_SCHEMA)
            # Remove invalid tis
            data["tis"] = [x for x in data["tis"] if x.get("name")]

        # Do the same for extra images, if we have them in this data structure
        if data.get("extra_source_images"):
            for i, img in enumerate(data.get("extra_source_images")):
                data["extra_source_images"][i] = self._validate_data_structure(img, HordeLib.EXTRA_IMAGES_SCHEMA)
            data["extra_source_images"] = [x for x in data["extra_source_images"] if x.get("image")]

        # Do the same for extra texts, if we have them in this data structure
        if data.get("extra_texts"):
            for i, img in enumerate(data.get("extra_texts")):
                data["extra_texts"][i] = self._validate_data_structure(img, HordeLib.EXTRA_TEXTS_SCHEMA)
            data["extra_texts"] = [x for x in data["extra_texts"] if x.get("text")]

        return data

    def _apply_aihorde_compatibility_hacks(self, payload: dict) -> tuple[dict, list[GenMetadataEntry]]:
        """For use by the AI Horde worker we require various counterintuitive hacks to the payload data.

        We encapsulate all of this implicit witchcraft in one function, here.
        """
        faults: list[GenMetadataEntry] = []

        if SharedModelManager.manager.compvis is None:
            raise RuntimeError("Cannot use AI Horde compatibility hacks without compvis loaded!")

        payload = deepcopy(payload)

        model = payload.get("model")

        if model is None:
            raise RuntimeError("No model specified in payload")

        # This is translated to "horde_model_name" later for compvis models and used as is for post processors
        payload["model_name"] = model

        found_model_in_ref = False
        found_model_on_disk = False
        model_files: list[dict] = [{}]

        if model in SharedModelManager.manager.compvis.model_reference:
            found_model_in_ref = True

        if SharedModelManager.manager.compvis.is_model_available(model):
            model_files = SharedModelManager.manager.compvis.get_model_filenames(model)
            found_model_on_disk = True

            if SharedModelManager.manager.compvis.model_reference[model].get("inpainting") is True:
                if payload.get("source_processing") not in ["inpainting", "outpainting"]:
                    logger.warning(
                        "Inpainting model detected, but source processing not set to inpainting or outpainting.",
                    )

                    payload["source_processing"] = "inpainting"

                source_image = payload.get("source_image")
                source_mask = payload.get("source_mask")

                if source_image is None or not isinstance(source_image, Image.Image):
                    logger.warning(
                        "Inpainting model detected, but source image is not a valid image. Using a noise image.",
                    )
                    faults.append(
                        GenMetadataEntry(
                            type=METADATA_TYPE.source_image,
                            value=METADATA_VALUE.parse_failed,
                        ),
                    )
                    payload["source_image"] = ImageUtils.create_noise_image(
                        payload["width"],
                        payload["height"],
                    )

                source_image = payload.get("source_image")

                if source_mask is None and (
                    source_image is None
                    or (isinstance(source_image, Image.Image) and not ImageUtils.has_alpha_channel(source_image))
                ):
                    logger.warning(
                        "Inpainting model detected, but no source mask provided. Using an all white mask.",
                    )
                    faults.append(
                        GenMetadataEntry(
                            type=METADATA_TYPE.source_mask,
                            value=METADATA_VALUE.parse_failed,
                        ),
                    )
                    payload["source_mask"] = ImageUtils.create_white_image(
                        source_image.width if source_image else payload["width"],
                        source_image.height if source_image else payload["height"],
                    )

        else:
            # The node may be a post processor, so we check the other model managers
            post_processor_model_managers = SharedModelManager.manager.get_model_manager_instances(
                [MODEL_CATEGORY_NAMES.codeformer, MODEL_CATEGORY_NAMES.esrgan, MODEL_CATEGORY_NAMES.gfpgan],
            )

            for post_processor_model_manager in post_processor_model_managers:
                if model in post_processor_model_manager.model_reference:
                    found_model_in_ref = True
                if post_processor_model_manager.is_model_available(model):
                    model_files = post_processor_model_manager.get_model_filenames(model)
                    found_model_on_disk = True
                    break

        if not found_model_in_ref:
            raise RuntimeError(f"Model {model} not found in model reference!")

        if not found_model_on_disk:
            raise RuntimeError(f"Model {model} not found on disk!")

        if len(model_files) == 0 or (not isinstance(model_files[0], dict)) or "file_path" not in model_files[0]:
            raise RuntimeError(f"Model {model} has no files in its reference entry!")

        payload["model"] = model_files[0]["file_path"]
        for file_entry in model_files:
            if "file_type" in file_entry:
                payload[file_entry["file_type"]] = file_entry["file_path"]

        # Rather than specify a scheduler, only karras or not karras is specified
        if payload.get("karras", False):
            payload["scheduler"] = "karras"
        else:
            payload["scheduler"] = "normal"

        prompt = payload.get("prompt")

        # Negative and positive prompts are merged together
        if prompt is not None:
            if "###" in prompt:
                split_prompts = prompt.split("###")
                payload["prompt"] = split_prompts[0]
                payload["negative_prompt"] = split_prompts[1]
        elif prompt == "":
            logger.warning("Empty prompt detected, this is likely to produce poor results")

        # Turn off hires fix if we're not generating a hires image, or if the params are just confused
        try:
            if "hires_fix" in payload:
                if SharedModelManager.manager.compvis.model_reference[model].get(
                    "baseline",
                ) == "stable diffusion 1" and (payload["width"] <= 512 or payload["height"] <= 512):
                    payload["hires_fix"] = False
                elif SharedModelManager.manager.compvis.model_reference[model].get(
                    "baseline",
                ) == "stable_diffusion_xl" and (payload["width"] <= 1024 or payload["height"] <= 1024):
                    payload["hires_fix"] = False
        except (TypeError, KeyError):
            payload["hires_fix"] = False

        # Turn off hires fix if we're inpainting as the dimensions are from the source image
        if "hires_fix" in payload and (
            payload.get("source_processing") == "inpainting" or payload.get("source_processing") == "outpainting"
        ):
            payload["hires_fix"] = False

        # Use denoising strength for both samplers if no second denoiser specified
        # but not for txt2img where denoising will always generally be 1.0
        if payload.get("hires_fix"):
            if payload.get("source_processing") and payload.get("source_processing") != "txt2img":
                if not payload.get("hires_fix_denoising_strength"):
                    payload["hires_fix_denoising_strength"] = payload.get("denoising_strength")

        if payload.get("workflow") == "qr_code":
            if payload.get("source_processing") and payload.get("source_processing") != "txt2img":
                if not payload.get("hires_fix_denoising_strength"):
                    payload["hires_fix_denoising_strength"] = payload.get("denoising_strength")

        # # Remap "denoising" to "controlnet strength", historical hack
        # if payload.get("control_type"):
        #     if payload.get("denoising_strength"):
        #         if not payload.get("control_strength"):
        #             payload["control_strength"] = payload["denoising_strength"]
        #             del payload["denoising_strength"]
        #         else:
        #             del payload["denoising_strength"]
        return payload, faults

    def _final_pipeline_adjustments(self, payload, pipeline_data) -> tuple[dict, list[GenMetadataEntry]]:
        payload = deepcopy(payload)
        faults: list[GenMetadataEntry] = []

        # Process dynamic prompts
        if new_prompt := DynamicPromptParser(payload["seed"]).parse(payload.get("prompt", "")):
            payload["prompt"] = new_prompt

        # Clip skip inversion, comfy uses -1, -2, etc
        if payload.get("clip_skip", 0) > 0:
            payload["clip_skip"] = -payload["clip_skip"]

        # Remap sampler name, horde to comfy
        payload["sampler_name"] = HordeLib.SAMPLERS_MAP.get(payload["sampler_name"])

        # Remap controlnet models, horde to comfy
        if payload.get("control_type"):
            payload["control_type"] = HordeLib.CONTROLNET_IMAGE_PREPROCESSOR_MAP.get(payload.get("control_type"))
        if payload.get("tis") and SharedModelManager.manager.ti:
            # Remove any requested TIs that we don't have
            for ti in payload.get("tis"):
                # Determine the actual ti filename
                if not SharedModelManager.manager.ti.is_local_model(str(ti["name"])):
                    try:
                        adhoc_ti = SharedModelManager.manager.ti.fetch_adhoc_ti(str(ti["name"]))
                    except Exception as e:
                        logger.info(f"Error fetching adhoc TI {ti['name']}: ({type(e).__name__}) {e}")
                        faults.append(
                            GenMetadataEntry(
                                type=METADATA_TYPE.ti,
                                value=METADATA_VALUE.download_failed,
                                ref=ti["name"],
                            ),
                        )
                        adhoc_ti = None
                    if not adhoc_ti:
                        logger.info(f"Adhoc TI requested '{ti['name']}' could not be found in CivitAI. Ignoring!")
                        faults.append(
                            GenMetadataEntry(
                                type=METADATA_TYPE.ti,
                                value=METADATA_VALUE.download_failed,
                                ref=ti["name"],
                            ),
                        )
                        continue
                ti_name = SharedModelManager.manager.ti.get_ti_name(str(ti["name"]))
                if ti_name:
                    logger.debug(f"Found valid TI {ti_name}")
                    if SharedModelManager.manager.compvis is None:
                        raise RuntimeError("Cannot use TIs without compvis loaded!")
                    model_details = SharedModelManager.manager.compvis.get_model_reference_info(payload["model"])
                    # If the ti and model do not match baseline, we ignore the TI
                    if not SharedModelManager.manager.ti.do_baselines_match(ti_name, model_details):
                        logger.info(f"Skipped TI {ti_name} because its baseline does not match the model's")
                        faults.append(
                            GenMetadataEntry(
                                type=METADATA_TYPE.ti,
                                value=METADATA_VALUE.baseline_mismatch,
                                ref=ti_name,
                            ),
                        )
                        continue
                    ti_inject = ti.get("inject_ti")
                    ti_strength = ti.get("strength", 1.0)
                    if type(ti_strength) not in [float, int]:
                        ti_strength = 1.0
                    ti_id = SharedModelManager.manager.ti.get_ti_id(str(ti["name"]))
                    if ti_inject == "prompt":
                        payload["prompt"] = f'(embedding:{ti_id}:{ti_strength}),{payload["prompt"]}'
                    elif ti_inject == "negprompt":
                        # create negative prompt if empty
                        if "negative_prompt" not in payload:
                            payload["negative_prompt"] = ""

                        had_leading_comma = payload["negative_prompt"].startswith(",")

                        payload["negative_prompt"] = f'{payload["negative_prompt"]},(embedding:{ti_id}:{ti_strength})'
                        if not had_leading_comma:
                            payload["negative_prompt"] = payload["negative_prompt"].strip(",")
                    SharedModelManager.manager.ti.touch_ti(ti_name)
        # Setup controlnet if required

        # For LORAs we completely build the LORA section of the pipeline dynamically, as we have
        # to handle n LORA models which form chained nodes in the pipeline.
        # Note that we build this between several nodes, the model_loader, clip_skip and the sampler,
        # plus the upscale sampler (used in hires fix) if there is one
        if payload.get("loras") and SharedModelManager.manager.lora:
            # Remove any requested LORAs that we don't have
            valid_loras = []
            for lora in payload.get("loras"):
                # Determine the actual lora filename
                is_version: bool = lora.get("is_version", False)
                verstext = ""
                if is_version:
                    verstext = " version"
                if not SharedModelManager.manager.lora.is_lora_available(str(lora["name"]), is_version=is_version):
                    logger.debug(f"Adhoc lora requested '{lora['name']}' not yet downloaded. Downloading...")
                    try:
                        adhoc_lora = SharedModelManager.manager.lora.fetch_adhoc_lora(
                            str(lora["name"]),
                            is_version=is_version,
                        )
                    except Exception as e:
                        logger.info(f"Error fetching adhoc lora {lora['name']}: ({type(e).__name__}) {e}")
                        faults.append(
                            GenMetadataEntry(
                                type=METADATA_TYPE.lora,
                                value=METADATA_VALUE.download_failed,
                                ref=lora["name"],
                            ),
                        )
                        adhoc_lora = None
                    if not adhoc_lora:
                        logger.info(
                            f"Adhoc lora requested{verstext} '{lora['name']} "
                            "could not be found in CivitAI. Ignoring!",
                        )
                        faults.append(
                            GenMetadataEntry(
                                type=METADATA_TYPE.lora,
                                value=METADATA_VALUE.download_failed,
                                ref=lora["name"],
                            ),
                        )
                        continue
                # We store the actual lora name to search for the trigger
                # If a version is requested, the lora name we need is the exact version
                if is_version:
                    lora_name = str(lora["name"])
                else:
                    lora_name = SharedModelManager.manager.lora.get_lora_name(str(lora["name"]))
                if lora_name is None:
                    faults.append(
                        GenMetadataEntry(
                            type=METADATA_TYPE.lora,
                            value=METADATA_VALUE.download_failed,
                            ref=lora_name,
                        ),
                    )
                    continue
                logger.debug(f"Found valid lora{verstext} {lora_name}")
                if SharedModelManager.manager.compvis is None:
                    raise RuntimeError("Cannot use LORAs without a compvis loaded!")
                model_details = SharedModelManager.manager.compvis.get_model_reference_info(payload["model"])
                # If the lora and model do not match baseline, we ignore the lora
                if not SharedModelManager.manager.lora.do_baselines_match(
                    lora_name,
                    model_details,
                    is_version=is_version,
                ):
                    logger.info(
                        f"Skipped lora{verstext} {lora_name} because its baseline does not match the model's",
                    )
                    faults.append(
                        GenMetadataEntry(
                            type=METADATA_TYPE.lora,
                            value=METADATA_VALUE.baseline_mismatch,
                            ref=lora_name,
                        ),
                    )
                    continue
                trigger_inject = lora.get("inject_trigger")
                trigger = None
                if trigger_inject == "any":
                    triggers = SharedModelManager.manager.lora.get_lora_triggers(lora_name, is_version=is_version)
                    if triggers:
                        trigger = random.choice(triggers)
                elif trigger_inject == "all":
                    triggers = SharedModelManager.manager.lora.get_lora_triggers(lora_name, is_version=is_version)
                    if triggers:
                        trigger = ", ".join(triggers)
                elif trigger_inject is not None:
                    trigger = SharedModelManager.manager.lora.find_lora_trigger(
                        lora_name,
                        trigger_inject,
                        is_version,
                    )
                if trigger:
                    # We inject at the start, to avoid throwing it in a negative prompt
                    payload["prompt"] = f'{trigger}, {payload["prompt"]}'
                # the fixed up and validated filename (Comfy expect the "name" key to be the filename)
                lora["name"] = SharedModelManager.manager.lora.get_lora_filename(lora_name, is_version=is_version)
                SharedModelManager.manager.lora._touch_lora(lora_name, is_version=is_version)
                valid_loras.append(lora)
            payload["loras"] = valid_loras
            for lora_index, lora in enumerate(payload.get("loras")):
                # Inject a lora node (first lora)
                if lora_index == 0:
                    pipeline_data[f"lora_{lora_index}"] = {
                        "inputs": {
                            "model": ["model_loader", 0],
                            "clip": ["model_loader", 1],
                            "lora_name": lora["name"],
                            "strength_model": lora["model"],
                            "strength_clip": lora["clip"],
                            # "model_manager": SharedModelManager,
                        },
                        "class_type": "HordeLoraLoader",
                    }
                else:
                    # Subsequent chained loras
                    pipeline_data[f"lora_{lora_index}"] = {
                        "inputs": {
                            "model": [f"lora_{lora_index-1}", 0],
                            "clip": [f"lora_{lora_index-1}", 1],
                            "lora_name": lora["name"],
                            "strength_model": lora["model"],
                            "strength_clip": lora["clip"],
                            # "model_manager": SharedModelManager,
                        },
                        "class_type": "HordeLoraLoader",
                    }

            for lora_index in range(len(payload.get("loras"))):
                # The first LORA always connects to the model loader
                if lora_index == 0:
                    self.generator.reconnect_input(pipeline_data, "lora_0.model", "model_loader")
                    self.generator.reconnect_input(pipeline_data, "lora_0.clip", "model_loader")
                else:
                    # Other loras connect to the previous lora
                    self.generator.reconnect_input(
                        pipeline_data,
                        f"lora_{lora_index}.model",
                        f"lora_{lora_index-1}.model",
                    )
                    self.generator.reconnect_input(
                        pipeline_data,
                        f"lora_{lora_index}.clip",
                        f"lora_{lora_index-1}.clip",
                    )

                # The last LORA always connects to the sampler and clip text encoders (via the clip_skip)
                if lora_index == len(payload.get("loras")) - 1 and SharedModelManager.manager.compvis:
                    model_details = SharedModelManager.manager.compvis.get_model_reference_info(payload["model_name"])
                    if model_details is not None and model_details["baseline"] == "flux_1":
                        self.generator.reconnect_input(pipeline_data, "cfg_guider.model", f"lora_{lora_index}")
                        self.generator.reconnect_input(pipeline_data, "basic_scheduler.model", f"lora_{lora_index}")
                    else:
                        self.generator.reconnect_input(pipeline_data, "sampler.model", f"lora_{lora_index}")
                        self.generator.reconnect_input(pipeline_data, "upscale_sampler.model", f"lora_{lora_index}")
                        self.generator.reconnect_input(pipeline_data, "clip_skip.clip", f"lora_{lora_index}")

        # Translate the payload parameters into pipeline parameters
        pipeline_params = {}
        for newkey, key in HordeLib.PAYLOAD_TO_PIPELINE_PARAMETER_MAPPING.items():
            multiplier = None
            # We allow a multiplier in the param, so that I can adjust easily the
            # values for steps on things like stable cascade
            if isinstance(key, FunctionType):
                pipeline_params[newkey] = key(payload)
            elif not isinstance(key, str):
                logger.error(f"Invalid key {key}")
                raise RuntimeError(f"Invalid key {key}")
            elif "*" in key:
                key, multiplier = key.split("*", 1)

            if key in payload:
                if multiplier:
                    pipeline_params[newkey] = round(payload.get(key) * float(multiplier))
                else:
                    pipeline_params[newkey] = payload.get(key)
            elif not isinstance(key, FunctionType):
                logger.error(f"Parameter {key} not found")

        # We inject these parameters to ensure the HordeCheckpointLoader knows what file to load, if necessary
        # We don't want to hardcode this into the pipeline.json as we export this directly from ComfyUI
        # and don't want to have to rememeber to re-add those keys
        if "model_loader_stage_c.ckpt_name" in pipeline_params:
            pipeline_params["model_loader_stage_c.file_type"] = "stable_cascade_stage_c"
        if "model_loader_stage_b.ckpt_name" in pipeline_params:
            pipeline_params["model_loader_stage_b.file_type"] = "stable_cascade_stage_b"
        pipeline_params["model_loader.file_type"] = None  # To allow normal SD pipelines to keep working

        # Inject our model manager
        # pipeline_params["model_loader.model_manager"] = SharedModelManager
        pipeline_params["model_loader.will_load_loras"] = bool(payload.get("loras"))
        pipeline_params["model_loader_stage_c.will_load_loras"] = False  # FIXME: Once we support loras
        # Does this have to be required var in the modelloader?
        pipeline_params["model_loader_stage_c.seamless_tiling_enabled"] = False
        pipeline_params["model_loader_stage_b.will_load_loras"] = False  # FIXME: Once we support loras
        # Does this have to be required var in the modelloader?
        pipeline_params["model_loader_stage_b.seamless_tiling_enabled"] = False

        # For hires fix, change the image sizes as we create an intermediate image first
        if payload.get("hires_fix", False):
            model_details = (
                SharedModelManager.manager.compvis.get_model_reference_info(payload["model_name"])
                if SharedModelManager.manager.compvis
                else None
            )

            original_width = pipeline_params.get("empty_latent_image.width")
            original_height = pipeline_params.get("empty_latent_image.height")

            if original_width is None or original_height is None:
                if model_details and model_details.get("baseline") == "stable diffusion 1":
                    logger.error("empty_latent_image.width or empty_latent_image.height not found. Using 512x512.")
                    original_width, original_height = (512, 512)
                else:
                    logger.error("empty_latent_image.width or empty_latent_image.height not found. Using 1024x1024.")
                    original_width, original_height = (1024, 1024)

            new_width, new_height = (None, None)
            baseline = None
            if model_details:
                baseline = model_details.get("baseline")
            if baseline:
                if baseline == "stable_cascade":
                    new_width, new_height = ImageUtils.get_first_pass_image_resolution_max(
                        original_width,
                        original_height,
                    )
                elif baseline == "stable_diffusion_xl":
                    new_width, new_height = ImageUtils.get_first_pass_image_resolution_sdxl(
                        original_width,
                        original_height,
                    )
                else:  # fall through case; only `stable diffusion 1`` at time of writing
                    new_width, new_height = ImageUtils.get_first_pass_image_resolution_min(
                        original_width,
                        original_height,
                    )

            # This is the *target* resolution
            pipeline_params["latent_upscale.width"] = original_width
            pipeline_params["latent_upscale.height"] = original_height

            if new_width and new_height:
                # This is the *first pass* resolution
                pipeline_params["empty_latent_image.width"] = new_width
                pipeline_params["empty_latent_image.height"] = new_height
            else:
                logger.error("Could not determine new image size for hires fix. Using 1024x1024.")
                pipeline_params["empty_latent_image.width"] = 1024
                pipeline_params["empty_latent_image.height"] = 1024

        if payload.get("control_type"):
            # Inject control net model manager
            # pipeline_params["controlnet_model_loader.model_manager"] = SharedModelManager
            model_name = self.CONTROLNET_MODEL_MAP.get(payload.get("control_type"))
            if not model_name:
                logger.error(f"Controlnet model for {payload.get('control_type')} not found")
            pipeline_params["controlnet_model_loader.control_net_name"] = model_name

            # Dynamically reconnect nodes in the pipeline to connect the correct pre-processor node
            if payload.get("return_control_map"):
                # Connect annotator to output image directly if we need to return the control map
                self.generator.reconnect_input(
                    pipeline_data,
                    "output_image.images",
                    payload["control_type"],
                )
            elif payload.get("image_is_control"):
                # Connect source image directly to controlnet apply node
                self.generator.reconnect_input(
                    pipeline_data,
                    "controlnet_apply.image",
                    "image_loader.image",
                )
            else:
                # Connect annotator to controlnet apply node
                self.generator.reconnect_input(
                    pipeline_data,
                    "controlnet_apply.image",
                    payload["control_type"],
                )

        # If we have a source image, use that rather than latent noise (i.e. img2img)
        # We do this by reconnecting the nodes in the pipeline to make the input to the vae encoder
        # the source image instead of the latent noise generator
        if pipeline_params.get("image_loader.image"):
            if SharedModelManager.manager.compvis:
                model_details = SharedModelManager.manager.compvis.get_model_reference_info(payload["model_name"])
                if isinstance(model_details, dict) and model_details.get("baseline") == "flux_1":
                    self.generator.reconnect_input(pipeline_data, "sampler_custom_advanced.latent_image", "vae_encode")
                else:
                    self.generator.reconnect_input(pipeline_data, "sampler.latent_image", "vae_encode")
        if pipeline_params.get("sc_image_loader.image"):
            self.generator.reconnect_input(
                pipeline_data,
                "sampler_stage_c.latent_image",
                "stablecascade_stagec_vaeencode",
            )
            self.generator.reconnect_input(
                pipeline_data,
                "sampler_stage_b.latent_image",
                "stablecascade_stagec_vaeencode",
            )

        # If we have a remix request, we check for extra images to add to the pipeline
        if payload.get("source_processing") == "remix":
            logger.debug([payload.get("source_image"), payload.get("extra_source_images")])
            for image_index in range(len(payload.get("extra_source_images", []))):
                # The first image is always taken from the source_image param
                # That will sit on the 0 spot.
                # Therefore we want the extra images to start iterating from 1
                extra_image = payload["extra_source_images"][image_index]["image"]
                extra_image_strength = payload["extra_source_images"][image_index].get("strength", 1)
                node_index = image_index + 1
                pipeline_data[f"sc_image_loader_{node_index}"] = {
                    "inputs": {"image": extra_image, "upload": "image"},
                    "class_type": "HordeImageLoader",
                }
                pipeline_data[f"clip_vision_encode_{node_index}"] = {
                    "inputs": {
                        "clip_vision": ["model_loader_stage_c", 3],
                        "image": [f"sc_image_loader_{node_index}", 0],
                        "crop": "center",
                    },
                    "class_type": "CLIPVisionEncode",
                }
                pipeline_data[f"unclip_conditioning_{node_index}"] = {
                    "inputs": {
                        "strength": extra_image_strength,
                        "noise_augmentation": 0,
                        # Each conditioning ingests the conditioning before it like a chain
                        "conditioning": [f"unclip_conditioning_{node_index-1}", 0],
                        "clip_vision_output": [f"clip_vision_encode_{node_index}", 0],
                    },
                    "class_type": "unCLIPConditioning",
                }

                # The last extra image always connects to the stage_c sampler positive prompt
                if image_index == len(payload.get("extra_source_images")) - 1:
                    self.generator.reconnect_input(
                        pipeline_data,
                        "sampler_stage_c.positive",
                        f"unclip_conditioning_{node_index}",
                    )

        # If we have a qr code request, we check for extra texts such as the generation url
        if payload.get("workflow") == "qr_code":
            original_width = pipeline_params.get("empty_latent_image.width", 512)
            original_height = pipeline_params.get("empty_latent_image.height", 512)
            pipeline_params["qr_code_split.max_image_size"] = max(original_width, original_height)
            pipeline_params["qr_code_split.text"] = "https://haidra.net"
            for text in payload.get("extra_texts"):
                if text["reference"] in ["qr_code", "qr_text"]:
                    pipeline_params["qr_code_split.text"] = text["text"]
                if text["reference"] == "protocol" and text["text"].lower() in ["https", "http"]:
                    pipeline_params["qr_code_split.protocol"] = text["text"].capitalize()
                if text["reference"] == "module_drawer" and text["text"].lower() in [
                    "square",
                    "gapped square",
                    "circle",
                    "rounded",
                    "vertical bars",
                    "horizontal bars",
                ]:
                    pipeline_params["qr_code_split.module_drawer"] = text["text"].capitalize()
                if text["reference"] == "function_layer_prompt":
                    pipeline_params["function_layer_prompt.text"] = text["text"]
                if text["reference"] == "x_offset" and text["text"].lstrip("-").isdigit():
                    x_offset = int(text["text"])
                    if x_offset < 0:
                        x_offset = 10
                    pipeline_params["qr_flattened_composite.x"] = x_offset
                if text["reference"] == "y_offset" and text["text"].lstrip("-").isdigit():
                    y_offset = int(text["text"])
                    if y_offset < 0:
                        y_offset = 10
                    pipeline_params["qr_flattened_composite.y"] = y_offset
                if text["reference"] == "qr_border" and text["text"].lstrip("-").isdigit():
                    border = int(text["text"])
                    if border < 0:
                        border = 10
                    pipeline_params["qr_code_split.border"] = border
            if not pipeline_params.get("qr_code_split.protocol"):
                pipeline_params["qr_code_split.protocol"] = "None"
            if not pipeline_params.get("function_layer_prompt.text"):
                pipeline_params["function_layer_prompt.text"] = payload["prompt"]
            try:
                test_qr = QRByModuleSizeSplitFunctionPatterns()
                _, _, _, _, _, qr_size = test_qr.generate_qr(
                    protocol=pipeline_params.get("qr_code_split.protocol"),
                    text=pipeline_params["qr_code_split.text"],
                    module_size=16,
                    max_image_size=pipeline_params["qr_code_split.max_image_size"],
                    fill_hexcolor="#FFFFFF",
                    back_hexcolor="#000000",
                    error_correction="High",
                    border=1,
                    module_drawer="Square",
                )
            except RuntimeError as err:
                logger.error(err)
                pipeline_params["qr_code_split.text"] = "This QR Code is too large for this image"
                test_qr = QRByModuleSizeSplitFunctionPatterns()
                qr_size = 624

            if not pipeline_params.get("qr_flattened_composite.x"):
                x_offset = int((original_width / 2) - qr_size / 2)
                # I don't know why but through trial and error I've discovered that the QR codes
                # are more legible when they're placed in an offset which is a multiple of 64
                x_offset = x_offset - (x_offset % 64) if x_offset % 64 != 0 else x_offset
                pipeline_params["qr_flattened_composite.x"] = x_offset
            if pipeline_params.get("qr_flattened_composite.x", 0) > int((original_width) - qr_size):
                pipeline_params["qr_flattened_composite.x"] = int((original_width) - qr_size) - 10
            if not pipeline_params.get("qr_flattened_composite.y"):
                y_offset = int((original_height / 2) - qr_size / 2)
                y_offset = y_offset - (y_offset % 64) if y_offset % 64 != 0 else y_offset
                pipeline_params["qr_flattened_composite.y"] = y_offset
            if pipeline_params.get("qr_flattened_composite.y", 0) > int((original_height) - qr_size):
                pipeline_params["qr_flattened_composite.y"] = int((original_height) - qr_size) - 10
            pipeline_params["module_layer_composite.x"] = pipeline_params["qr_flattened_composite.x"]
            pipeline_params["module_layer_composite.y"] = pipeline_params["qr_flattened_composite.y"]
            pipeline_params["function_layer_composite.x"] = pipeline_params["qr_flattened_composite.x"]
            pipeline_params["function_layer_composite.y"] = pipeline_params["qr_flattened_composite.y"]
            pipeline_params["mask_composite.x"] = pipeline_params["qr_flattened_composite.x"]
            pipeline_params["mask_composite.y"] = pipeline_params["qr_flattened_composite.y"]
            if SharedModelManager.manager.compvis:
                model_details = SharedModelManager.manager.compvis.get_model_reference_info(payload["model_name"])
                if model_details and model_details.get("baseline") == "stable diffusion 1":
                    pipeline_params["controlnet_qr_model_loader.control_net_name"] = (
                        "control_v1p_sd15_qrcode_monster_v2.safetensors"
                    )
        if payload.get("transparent") is True:
            # A transparent gen is basically a fancy lora
            pipeline_params["model_loader.will_load_loras"] = True
            if SharedModelManager.manager.compvis:
                model_details = SharedModelManager.manager.compvis.get_model_reference_info(payload["model_name"])
                # SD2, Cascade and SD3 not supported
                if model_details and model_details.get("baseline") in ["stable diffusion 1", "stable_diffusion_xl"]:
                    self.generator.reconnect_input(pipeline_data, "sampler.model", "layer_diffuse_apply")
                    self.generator.reconnect_input(pipeline_data, "layer_diffuse_apply.model", "model_loader")
                    self.generator.reconnect_input(pipeline_data, "output_image.images", "layer_diffuse_decode_rgba")
                    self.generator.reconnect_input(pipeline_data, "layer_diffuse_decode_rgba.images", "vae_decode")
                    if payload.get("hires_fix") is True:
                        self.generator.reconnect_input(pipeline_data, "upscale_sampler.model", "layer_diffuse_apply")
                    if model_details.get("baseline") == "stable diffusion 1":
                        pipeline_params["layer_diffuse_apply.config"] = "SD15, Attention Injection, attn_sharing"
                        pipeline_params["layer_diffuse_decode_rgba.sd_version"] = "SD15"
                    else:
                        pipeline_params["layer_diffuse_apply.config"] = "SDXL, Conv Injection"
                        pipeline_params["layer_diffuse_decode_rgba.sd_version"] = "SDXL"

        return pipeline_params, faults

    def _get_appropriate_pipeline(self, params):
        # Determine the correct pipeline based on the parameters we have
        #
        # The pipelines are:
        #
        #     stable_diffusion
        #       stable_diffusion_hires_fix
        #     stable_diffusion_img2img_mask
        #     stable_diffusion_paint
        #     controlnet
        #       controlnet_hires_fix
        #       controlnet_annotator
        #     image_facefix
        #     image_upscale
        #     stable_cascade
        #       stable_cascade_remix
        #       stable_cascade_2pass
        #     qr_code

        # controlnet, controlnet_hires_fix controlnet_annotator
        if params.get("workflow") == "qr_code":
            return "qr_code"
        if params.get("model_name"):
            model_details = SharedModelManager.manager.compvis.get_model_reference_info(params["model_name"])
            if model_details.get("baseline") == "stable_cascade":
                if params.get("source_processing") == "remix":
                    return "stable_cascade_remix"
                if params.get("hires_fix", False):
                    return "stable_cascade_2pass"
                return "stable_cascade"
            if model_details.get("baseline") == "flux_1":
                return "flux"
        if params.get("control_type"):
            if params.get("return_control_map", False):
                return "controlnet_annotator"

            if params.get("hires_fix"):
                return "controlnet_hires_fix"

            return "controlnet"

        # stable_diffusion_paint, stable_diffusion_img2img_mask
        if params.get("source_processing") == "img2img":
            has_mask = params.get("source_mask") or (
                params.get("source_image") and len(params.get("source_image", "").getbands()) == 4
            )
            if has_mask:
                return "stable_diffusion_img2img_mask"
        elif params.get("source_processing") == "inpainting":
            return "stable_diffusion_paint"
        elif params.get("source_processing") == "outpainting":
            return "stable_diffusion_paint"

        # stable_diffusion and stable_diffusion_hires_fix
        if params.get("hires_fix", False):
            return "stable_diffusion_hires_fix"

        return "stable_diffusion"  # also includes img2img mode

    def _process_results(
        self,
        images: list[dict[str, typing.Any]],
    ) -> list[tuple[Image.Image, io.BytesIO]]:
        if images is None:
            logger.error("No images returned from comfy pipeline!")
            raise RuntimeError("No images returned from comfy pipeline!")

        results: list[tuple[Image.Image, io.BytesIO]] = []

        # Return image(s) or raw PNG byte stream
        for image in images:
            results.append(
                (
                    Image.open(image["imagedata"]),
                    image["imagedata"],
                ),
            )
        return results

    def _get_validated_payload_and_pipeline_data(self, payload: dict) -> tuple[dict, dict, list[GenMetadataEntry]]:
        # AIHorde hacks to payload
        payload, compatibility_faults = self._apply_aihorde_compatibility_hacks(payload)
        # Check payload types/values and normalise it's format
        payload = self._validate_data_structure(payload)
        # Resize the source image and mask to actual final width/height requested
        ImageUtils.resize_sources_to_request(payload)
        # Determine the correct pipeline to use
        pipeline = self._get_appropriate_pipeline(payload)
        # Final adjustments to the pipeline
        pipeline_data = self.generator.get_pipeline_data(pipeline)
        payload, finale_adjustment_faults = self._final_pipeline_adjustments(payload, pipeline_data)
        return payload, pipeline_data, compatibility_faults + finale_adjustment_faults

    def _inference(
        self,
        payload: dict,
        *,
        single_image_expected: bool = True,
        comfyui_progress_callback: Callable[[ComfyUIProgress, str], None] | None = None,
    ) -> list[ResultingImageReturn] | ResultingImageReturn:
        payload, pipeline_data, faults = self._get_validated_payload_and_pipeline_data(payload)

        # Run the pipeline

        # Add prefix to loras to avoid name collisions with other models
        loras_being_used = [f"lora-{x['name']}" for x in payload.get("loras", []) if x]
        tis_being_used = [f"ti-{x['name']}" for x in payload.get("tis", []) if x]

        # main model
        main_model = f"{payload.get('model_loader.horde_model_name')}"
        # controlnet model
        controlnet_name = payload.get("controlnet_model_loader.control_net_name")

        resolution = f"{payload.get('empty_latent_image.width')}x{payload.get('empty_latent_image.height')}"
        steps = payload.get("sampler.steps")

        logger.info(f"Generating image {resolution} in size for {steps} steps with model {main_model}.")
        if "latent_upscale.width" in payload:
            hires_fix_final_resolution = (
                f"{payload.get('latent_upscale.width')}x{payload.get('latent_upscale.height')}"
            )
            logger.info(f"Using hiresfix, final resolution {hires_fix_final_resolution}.")
        if controlnet_name:
            logger.info(f"Using controlnet {controlnet_name}")
        if loras_being_used:
            logger.info(f"Using {len(loras_being_used)} LORAs")
        if tis_being_used:
            logger.info(f"Using {len(tis_being_used)} TIs")

        # Call the inference pipeline
        # logger.debug(payload)
        images = self.generator.run_image_pipeline(pipeline_data, payload, comfyui_progress_callback)

        results = self._process_results(images)
        ret_results = [
            ResultingImageReturn(
                image=image,
                rawpng=rawpng,
                faults=faults,
            )
            for image, rawpng in results
        ]

        if single_image_expected:
            if len(results) != 1:
                raise RuntimeError("Expected a single image but got multiple")
            return ret_results[0]

        return ret_results

    def basic_inference(
        self,
        payload: dict | ImageGenerateJobPopResponse,
        *,
        progress_callback: Callable[[ProgressReport], None] | None = None,
    ) -> list[ResultingImageReturn]:
        post_processing_requested: list[str] | None = None
        if isinstance(payload, dict):
            post_processing_requested = payload.get("post_processing")

        faults = []
        if isinstance(payload, ImageGenerateJobPopResponse):  # TODO move this to _inference()
            for post_processor_requested in payload.payload.post_processing:
                if post_processing_requested is None:
                    post_processing_requested = []
                post_processing_requested.append(post_processor_requested)
                logger.debug(f"Post-processing requested: {post_processor_requested}")

            sub_payload = payload.payload.model_dump()

            def handle_images(
                payload: ImageGenerateJobPopResponse,
                image_type: str,
                get_downloaded_image_func: Callable,
            ):
                image = getattr(payload, image_type)

                if image is not None and "http" in image:
                    image = get_downloaded_image_func()

                    if image is None:
                        logger.error(
                            f"{image_type.capitalize()} is a URL but wasn't downloaded, "
                            "this is not supported in this context. Run the `async_download_*` methods first.",
                        )

                        return None

                return image

            source_image = handle_images(
                payload,
                "source_image",
                payload.get_downloaded_source_image,
            )
            if source_image is None:
                logger.info("No source image found in payload.")

            mask_image = handle_images(
                payload,
                "source_mask",
                payload.get_downloaded_source_mask,
            )
            if mask_image is None:
                logger.info("No mask image found in payload.")

            extra_source_images = payload.extra_source_images

            if extra_source_images is not None:
                extra_source_images = payload.get_downloaded_extra_source_images()
                if extra_source_images is not None:
                    logger.info(f"Using {len(extra_source_images)} downloaded extra source images.")
                else:
                    logger.info("No extra source images found in payload.")

            esi_to_remove = []
            if extra_source_images is not None:
                for esi in extra_source_images:
                    if "http" in esi.image:
                        logger.warning("Extra source image is a URL, this is not supported in this context.")
                        esi_to_remove.append(esi)

                extra_source_images = [esi for esi in extra_source_images if esi not in esi_to_remove]
            # If its a base64 encoded image, decode it
            if isinstance(source_image, str):
                try:
                    source_image_bytes = base64.b64decode(source_image)
                    source_image_pil = Image.open(io.BytesIO(source_image_bytes))
                    sub_payload["source_image"] = source_image_pil
                except Exception as err:
                    faults.append(
                        GenMetadataEntry(
                            type=METADATA_TYPE.source_image,
                            value=METADATA_VALUE.parse_failed,
                        ),
                    )
                    logger.warning(f"Failed to parse source image ({err}). Falling back to text2img.")

            if isinstance(mask_image, str):
                try:
                    mask_image_bytes = base64.b64decode(mask_image)
                    mask_image_pil = Image.open(io.BytesIO(mask_image_bytes))
                    sub_payload["source_mask"] = mask_image_pil
                except Exception as err:
                    faults.append(
                        GenMetadataEntry(
                            type=METADATA_TYPE.source_mask,
                            value=METADATA_VALUE.parse_failed,
                        ),
                    )
                    logger.warning(f"Failed to parse source mask ({err}). Ignoring it.")

            if isinstance(extra_source_images, list):
                extra_source_images_sub = []
                for esi_index, esi in enumerate(extra_source_images):
                    try:
                        esi_bytes = base64.b64decode(esi.image)
                        esi_pil = Image.open(io.BytesIO(esi_bytes))
                        extra_source_images_sub.append(
                            {
                                "image": esi_pil,
                                "strength": esi.strength,
                            },
                        )
                    except Exception as err:
                        faults.append(
                            GenMetadataEntry(
                                type=METADATA_TYPE.extra_source_images,
                                value=METADATA_VALUE.parse_failed,
                                ref=str(esi_index),
                            ),
                        )
                        logger.warning(f"Failed to parse extra source image {esi_index} ({err}). Ignoring it.")
                sub_payload["extra_source_images"] = extra_source_images_sub

            sub_payload["source_processing"] = payload.source_processing
            sub_payload["model"] = payload.model
            payload = sub_payload

        if progress_callback is not None:
            try:
                progress_callback(
                    ProgressReport(
                        hordelib_progress_state=ProgressState.started,
                        hordelib_message="Initiating inference...",
                        progress=0,
                    ),
                )
            except Exception as e:
                logger.error(f"Progress callback failed ({type(e)}): {e}")

        def _default_progress_callback(comfyui_progress: ComfyUIProgress, message: str) -> None:
            nonlocal progress_callback
            if progress_callback is not None:
                try:
                    progress_callback(
                        ProgressReport(
                            hordelib_progress_state=ProgressState.progress,
                            hordelib_message=message,
                            comfyui_progress=comfyui_progress,
                        ),
                    )
                except Exception as e:
                    logger.error(f"Progress callback failed ({type(e)}): {e}")

        result = self._inference(
            payload,
            single_image_expected=False,
            comfyui_progress_callback=_default_progress_callback,
        )

        if not isinstance(result, list):
            raise RuntimeError(f"Expected a list of PIL.Image.Image but got {type(result)}")

        return_list = [x for x in result if isinstance(x.image, Image.Image)]
        pptext = ""
        if post_processing_requested is not None:
            pptext = " Initiating post-processing..."
        logger.debug(f"Inference complete. Received {len(return_list)} images.{pptext}")

        post_processed: list[ResultingImageReturn] | None = None
        if post_processing_requested is not None:
            if progress_callback is not None:
                try:
                    progress_callback(
                        ProgressReport(
                            hordelib_progress_state=ProgressState.post_processing,
                            hordelib_message="Post Processing.",
                        ),
                    )
                except Exception as e:
                    logger.error(f"Progress callback failed ({type(e)}): {e}")

            post_processed = []
            for ret in return_list:
                single_image_faults = faults[:]
                final_image = ret.image
                final_rawpng = ret.rawpng

                if progress_callback is not None:
                    try:
                        progress_callback(
                            ProgressReport(
                                hordelib_progress_state=ProgressState.progress,
                                hordelib_message="Post Processing new image.",
                            ),
                        )
                    except Exception as e:
                        logger.error(f"Progress callback failed ({type(e)}): {e}")

                # Ensure facefixers always happen first
                post_processing_requested = sorted(
                    post_processing_requested,
                    key=lambda x: 1 if x in KNOWN_FACEFIXERS.__members__ else 0,
                )

                for post_processing in post_processing_requested:
                    if (
                        post_processing in KNOWN_UPSCALERS.__members__
                        or post_processing in KNOWN_UPSCALERS._value2member_map_
                    ):
                        image_ret = self.image_upscale(
                            {
                                "model": post_processing,
                                "source_image": final_image,
                            },
                        )
                        single_image_faults += image_ret.faults
                        final_rawpng = image_ret.rawpng
                        final_image = image_ret.image

                    elif post_processing in KNOWN_FACEFIXERS.__members__:
                        image_ret = self.image_facefix(
                            {
                                "model": post_processing,
                                "source_image": final_image,
                                "facefixer_strength": payload.get("facefixer_strength", 1.0),
                            },
                        )
                        single_image_faults += image_ret.faults
                        final_rawpng = image_ret.rawpng
                        final_image = image_ret.image

                    elif post_processing == "strip_background":
                        if final_image is not None:
                            final_image = ImageUtils.strip_background(final_image)

                if final_image is None:
                    # TODO: Allow to return a partially PP image?
                    logger.error("Post processing failed and there is no output image!")
                else:
                    post_processed.append(
                        ResultingImageReturn(image=final_image, rawpng=final_rawpng, faults=single_image_faults),
                    )

        if progress_callback is not None:
            try:
                progress_callback(
                    ProgressReport(
                        hordelib_progress_state=ProgressState.finished,
                        hordelib_message="Inference complete.",
                        progress=100,
                    ),
                )
            except Exception as e:
                logger.error(f"Progress callback failed ({type(e)}): {e}")

        if post_processed is not None:
            logger.debug(f"Post-processing complete. Returning {len(post_processed)} images.")
            return post_processed

        if len(return_list) == len(result):
            return return_list

        raise RuntimeError("Expected a list of PIL.Image.Image but got a mix of types!")

    def basic_inference_single_image(self, payload: dict) -> ResultingImageReturn:
        result = self._inference(payload, single_image_expected=True)
        if isinstance(result, ResultingImageReturn):
            return result

        raise RuntimeError(f"Expected a PIL.Image.Image but got {type(result)}")

    def basic_inference_rawpng(self, payload: dict) -> list[io.BytesIO]:
        """Return the results directly from comfy as (a) raw PNG byte stream(s)."""
        result = self._inference(payload, single_image_expected=False)

        if isinstance(result, list):
            bytes_io_list = [x.rawpng for x in result if isinstance(x.rawpng, io.BytesIO)]
            if len(bytes_io_list) == len(result):
                return bytes_io_list

            raise RuntimeError("Expected a list of io.BytesIO but got a mix of types!")

        if isinstance(result.image, io.BytesIO):
            return [result.image]

        raise RuntimeError(f"Expected at least one io.BytesIO. Got {result}.")

    def image_upscale(self, payload) -> ResultingImageReturn:
        logger.debug("image_upscale called")

        from hordelib.comfy_horde import log_free_ram

        log_free_ram()

        # AIHorde hacks to payload
        payload, compatibility_faults = self._apply_aihorde_compatibility_hacks(payload)
        # Remember if we were passed width and height, we wouldn't normally be passed width and height
        # because the upscale models upscale to a fixed multiple of image size. However, if we *are*
        # passed a width and height we rescale the upscale output image to this size.
        width = payload.get("height")
        height = payload.get("width")
        # Check payload types/values and normalise it's format
        payload = self._validate_data_structure(payload)
        # Final adjustments to the pipeline
        pipeline_name = "image_upscale"
        pipeline_data = self.generator.get_pipeline_data(pipeline_name)
        payload, final_adjustment_faults = self._final_pipeline_adjustments(payload, pipeline_data)

        # Run the pipeline

        images = self.generator.run_image_pipeline(pipeline_data, payload)

        # Allow arbitrary resizing by shrinking the image back down
        if width or height:
            return ResultingImageReturn(
                ImageUtils.shrink_image(Image.open(images[0]["imagedata"]), width, height),
                rawpng=None,
                faults=final_adjustment_faults,
            )
        result = self._process_results(images)
        if len(result) != 1:
            raise RuntimeError("Expected a single image but got multiple")

        image, rawpng = result[0]
        if not isinstance(image, Image.Image):
            raise RuntimeError(f"Expected a PIL.Image.Image but got {type(image)}")

        log_free_ram()
        return ResultingImageReturn(image=image, rawpng=rawpng, faults=compatibility_faults + final_adjustment_faults)

    def image_facefix(self, payload) -> ResultingImageReturn:
        logger.debug("image_facefix called")

        from hordelib.comfy_horde import log_free_ram

        log_free_ram()

        # AIHorde hacks to payload
        payload, compatibility_faults = self._apply_aihorde_compatibility_hacks(payload)
        # Check payload types/values and normalise it's format
        payload = self._validate_data_structure(payload)
        # Final adjustments to the pipeline
        pipeline_name = "image_facefix"
        pipeline_data = self.generator.get_pipeline_data(pipeline_name)
        payload, final_adjustment_faults = self._final_pipeline_adjustments(payload, pipeline_data)

        # Run the pipeline

        images = self.generator.run_image_pipeline(pipeline_data, payload)

        results = self._process_results(images)
        if len(results) != 1:
            raise RuntimeError("Expected a single image but got multiple")

        image, rawpng = results[0]
        if not isinstance(image, Image.Image):
            raise RuntimeError(f"Expected a PIL.Image.Image but got {type(image)}")

        log_free_ram()

        return ResultingImageReturn(image=image, rawpng=rawpng, faults=compatibility_faults + final_adjustment_faults)



# test_horde.py

import pytest
from PIL import Image

from hordelib.horde import HordeLib
from hordelib.shared_model_manager import SharedModelManager

from .testing_shared_functions import check_list_inference_images_similarity, check_single_inference_image_similarity


class TestHordeInferenceFlux:

    @pytest.mark.default_flux1_model
    def test_flux_schnell_fp8_text_to_image(
        self,
        hordelib_instance: HordeLib,
        flux1_schnell_fp8_base_model_name: str,
    ):
        data = {
            "sampler_name": "k_euler",
            "cfg_scale": 1,
            "denoising_strength": 1.0,
            "seed": 13122,
            "height": 1024,
            "width": 1024,
            "karras": False,
            "tiling": False,
            "hires_fix": False,
            "clip_skip": 1,
            "control_type": None,
            "image_is_control": False,
            "return_control_map": False,
            "prompt": 'a steampunk text that says "Horde Engine" floating',
            "ddim_steps": 4,
            "n_iter": 1,
            "model": flux1_schnell_fp8_base_model_name,
        }
        pil_image = hordelib_instance.basic_inference_single_image(data).image
        assert pil_image is not None
        assert isinstance(pil_image, Image.Image)

        img_filename = "flux_schnell_fp8_text_to_image.png"
        pil_image.save(f"images/{img_filename}", quality=100)

        assert check_single_inference_image_similarity(
            f"images_expected/{img_filename}",
            pil_image,
        )

    @pytest.mark.default_flux1_model
    def test_flux_schnell_fp8_text_to_image_n_iter(
        self,
        hordelib_instance: HordeLib,
        flux1_schnell_fp8_base_model_name: str,
    ):
        data = {
            "sampler_name": "k_euler",
            "cfg_scale": 1,
            "denoising_strength": 1.0,
            "seed": 13122,
            "height": 1024,
            "width": 1024,
            "karras": False,
            "tiling": False,
            "hires_fix": False,
            "clip_skip": 1,
            "control_type": None,
            "image_is_control": False,
            "return_control_map": False,
            "prompt": 'a steampunk text that says "Horde Engine" floating',
            "ddim_steps": 4,
            "n_iter": 2,
            "model": flux1_schnell_fp8_base_model_name,
        }
        image_results = hordelib_instance.basic_inference(data)

        assert len(image_results) == 2

        img_pairs_to_check = []

        img_filename_base = "flux_schnell_fp8_text_to_image_n_iter_{0}.png"

        for i, image_result in enumerate(image_results):
            assert image_result.image is not None
            assert isinstance(image_result.image, Image.Image)

            img_filename = img_filename_base.format(i)

            image_result.image.save(f"images/{img_filename}", quality=100)
            img_pairs_to_check.append((f"images_expected/{img_filename}", image_result.image))

        assert check_single_inference_image_similarity(
            "images_expected/flux_schnell_fp8_text_to_image.png",
            "images/flux_schnell_fp8_text_to_image_n_iter_0.png",
        )

        assert check_list_inference_images_similarity(img_pairs_to_check)

    @pytest.mark.default_flux1_model
    def test_flux_schnell_fp8_image_to_image(
        self,
        hordelib_instance: HordeLib,
        flux1_schnell_fp8_base_model_name: str,
    ):
        data = {
            "sampler_name": "k_euler",
            "cfg_scale": 1,
            "denoising_strength": 0.8,
            "seed": 13122,
            "height": 1024,
            "width": 1024,
            "karras": False,
            "tiling": False,
            "hires_fix": False,
            "clip_skip": 1,
            "control_type": None,
            "image_is_control": False,
            "return_control_map": False,
            "prompt": "a steampunk cowboy walking away from an explosion",
            "ddim_steps": 4,
            "n_iter": 1,
            "source_image": Image.open("images/test_db0.jpg"),
            "source_processing": "img2img",
            "model": flux1_schnell_fp8_base_model_name,
        }
        pil_image = hordelib_instance.basic_inference_single_image(data).image
        assert pil_image is not None
        assert isinstance(pil_image, Image.Image)

        img_filename = "flux_schnell_fp8_image_to_image.png"
        pil_image.save(f"images/{img_filename}", quality=100)

        assert check_single_inference_image_similarity(
            f"images_expected/{img_filename}",
            pil_image,
        )

    # @pytest.mark.default_flux1_model
    # def test_flux_dev_fp8_image_to_image(
    #     self,
    #     hordelib_instance: HordeLib,
    #     flux1_dev_fp8_base_model_name: str,
    # ):
    #     data = {
    #         "sampler_name": "k_euler",
    #         "cfg_scale": 2,
    #         "denoising_strength": 0.85,
    #         "seed": 13122,
    #         "height": 1024,
    #         "width": 1024,
    #         "karras": False,
    #         "tiling": False,
    #         "hires_fix": False,
    #         "clip_skip": 1,
    #         "control_type": None,
    #         "image_is_control": False,
    #         "return_control_map": False,
    #         "prompt": "a goth cowboy in black walking away from an explosion",
    #         "ddim_steps": 20,
    #         "n_iter": 1,
    #         "source_image": Image.open("images/test_db0.jpg"),
    #         "source_processing": "img2img",
    #         "model": flux1_dev_fp8_base_model_name,
    #     }
    #     pil_image = hordelib_instance.basic_inference_single_image(data).image
    #     assert pil_image is not None
    #     assert isinstance(pil_image, Image.Image)

    #     img_filename = "flux_dev_fp8_image_to_image.png"
    #     pil_image.save(f"images/{img_filename}", quality=100)

    #     assert check_single_inference_image_similarity(
    #         f"images_expected/{img_filename}",
    #         pil_image,
    #     )

    @pytest.mark.default_flux1_model
    def test_flux_dev_fp8_text_to_image_lora(
        self,
        shared_model_manager: type[SharedModelManager],
        hordelib_instance: HordeLib,
        flux1_schnell_fp8_base_model_name: str,
    ):

        lora_schnell_version_id = "812384"
        lora_dev_version_id = "735063"

        if shared_model_manager.manager.lora:
            assert shared_model_manager.manager.lora.fetch_adhoc_lora(
                lora_schnell_version_id,
                timeout=45,
                is_version=True,
            )

            assert shared_model_manager.manager.lora.fetch_adhoc_lora(
                lora_dev_version_id,
                timeout=45,
                is_version=True,
            )

        data = {
            "sampler_name": "k_euler",
            "cfg_scale": 1,
            "denoising_strength": 1.0,
            "seed": 13122,
            "height": 1024,
            "width": 1024,
            "karras": False,
            "tiling": False,
            "hires_fix": False,
            "clip_skip": 1,
            "control_type": None,
            "image_is_control": False,
            "return_control_map": False,
            "prompt": (
                "A close-up of an magical and ominous obsidian monolith, with an glowing orange cat print inside. "
                "detailmaximizer, dreamy art"
            ),
            "ddim_steps": 4,
            "n_iter": 1,
            "model": flux1_schnell_fp8_base_model_name,
            "loras": [
                {"name": lora_schnell_version_id, "model": 1.0, "clip": 1.0, "is_version": True},
                {"name": lora_dev_version_id, "model": 0.5, "clip": 1.0, "is_version": True},
            ],
        }
        pil_image = hordelib_instance.basic_inference_single_image(data).image
        assert pil_image is not None
        assert isinstance(pil_image, Image.Image)

        img_filename = "flux_schnell_fp8_text_to_image_lora.png"
        pil_image.save(f"images/{img_filename}", quality=100)

        assert check_single_inference_image_similarity(
            f"images_expected/{img_filename}",
            pil_image,
        )



from typing import Union

from hordelib.nodes.facerestore_cf.r_chainner.gfpganv1_clean_arch import GFPGANv1Clean


PyTorchFaceModels = (GFPGANv1Clean,)
PyTorchFaceModel = Union[GFPGANv1Clean]


def is_pytorch_face_model(model: object):
    return isinstance(model, PyTorchFaceModels)


PyTorchModels = (*PyTorchFaceModels,)
PyTorchModel = Union[PyTorchFaceModel]


def is_pytorch_model(model: object):
    return isinstance(model, PyTorchModels)



import comfy.samplers
import comfy.utils
import torch
import numpy as np
from tqdm.auto import trange


@torch.no_grad()
def sample_lcm_upscale(model, x, sigmas, extra_args=None, callback=None, disable=None, total_upscale=2.0, upscale_method="bislerp", upscale_steps=None):
    extra_args = {} if extra_args is None else extra_args

    if upscale_steps is None:
        upscale_steps = max(len(sigmas) // 2 + 1, 2)
    else:
        upscale_steps += 1
        upscale_steps = min(upscale_steps, len(sigmas) + 1)

    upscales = np.linspace(1.0, total_upscale, upscale_steps)[1:]

    orig_shape = x.size()
    s_in = x.new_ones([x.shape[0]])
    for i in trange(len(sigmas) - 1, disable=disable):
        denoised = model(x, sigmas[i] * s_in, **extra_args)
        if callback is not None:
            callback({'x': x, 'i': i, 'sigma': sigmas[i], 'sigma_hat': sigmas[i], 'denoised': denoised})

        x = denoised
        if i < len(upscales):
            x = comfy.utils.common_upscale(x, round(orig_shape[-1] * upscales[i]), round(orig_shape[-2] * upscales[i]), upscale_method, "disabled")

        if sigmas[i + 1] > 0:
            x += sigmas[i + 1] * torch.randn_like(x)
    return x


class SamplerLCMUpscale:
    upscale_methods = ["bislerp", "nearest-exact", "bilinear", "area", "bicubic"]

    @classmethod
    def INPUT_TYPES(s):
        return {"required":
                    {"scale_ratio": ("FLOAT", {"default": 1.0, "min": 0.1, "max": 20.0, "step": 0.01}),
                     "scale_steps": ("INT", {"default": -1, "min": -1, "max": 1000, "step": 1}),
                     "upscale_method": (s.upscale_methods,),
                      }
               }
    RETURN_TYPES = ("SAMPLER",)
    CATEGORY = "sampling/custom_sampling/samplers"

    FUNCTION = "get_sampler"

    def get_sampler(self, scale_ratio, scale_steps, upscale_method):
        if scale_steps < 0:
            scale_steps = None
        sampler = comfy.samplers.KSAMPLER(sample_lcm_upscale, extra_options={"total_upscale": scale_ratio, "upscale_steps": scale_steps, "upscale_method": upscale_method})
        return (sampler, )

from comfy.k_diffusion.sampling import to_d
import comfy.model_patcher

@torch.no_grad()
def sample_euler_pp(model, x, sigmas, extra_args=None, callback=None, disable=None):
    extra_args = {} if extra_args is None else extra_args

    temp = [0]
    def post_cfg_function(args):
        temp[0] = args["uncond_denoised"]
        return args["denoised"]

    model_options = extra_args.get("model_options", {}).copy()
    extra_args["model_options"] = comfy.model_patcher.set_model_options_post_cfg_function(model_options, post_cfg_function, disable_cfg1_optimization=True)

    s_in = x.new_ones([x.shape[0]])
    for i in trange(len(sigmas) - 1, disable=disable):
        sigma_hat = sigmas[i]
        denoised = model(x, sigma_hat * s_in, **extra_args)
        d = to_d(x - denoised + temp[0], sigmas[i], denoised)
        if callback is not None:
            callback({'x': x, 'i': i, 'sigma': sigmas[i], 'sigma_hat': sigma_hat, 'denoised': denoised})
        dt = sigmas[i + 1] - sigma_hat
        x = x + d * dt
    return x


class SamplerEulerCFGpp:
    @classmethod
    def INPUT_TYPES(s):
        return {"required":
                    {"version": (["regular", "alternative"],),}
               }
    RETURN_TYPES = ("SAMPLER",)
    # CATEGORY = "sampling/custom_sampling/samplers"
    CATEGORY = "_for_testing"

    FUNCTION = "get_sampler"

    def get_sampler(self, version):
        if version == "alternative":
            sampler = comfy.samplers.KSAMPLER(sample_euler_pp)
        else:
            sampler = comfy.samplers.ksampler("euler_cfg_pp")
        return (sampler, )

NODE_CLASS_MAPPINGS = {
    "SamplerLCMUpscale": SamplerLCMUpscale,
    "SamplerEulerCFGpp": SamplerEulerCFGpp,
}

NODE_DISPLAY_NAME_MAPPINGS = {
    "SamplerEulerCFGpp": "SamplerEulerCFG++",
}



import torch
from typing import Optional
import comfy.ldm.modules.diffusionmodules.mmdit

class ControlNet(comfy.ldm.modules.diffusionmodules.mmdit.MMDiT):
    def __init__(
        self,
        num_blocks = None,
        control_latent_channels = None,
        dtype = None,
        device = None,
        operations = None,
        **kwargs,
    ):
        super().__init__(dtype=dtype, device=device, operations=operations, final_layer=False, num_blocks=num_blocks, **kwargs)
        # controlnet_blocks
        self.controlnet_blocks = torch.nn.ModuleList([])
        for _ in range(len(self.joint_blocks)):
            self.controlnet_blocks.append(operations.Linear(self.hidden_size, self.hidden_size, device=device, dtype=dtype))

        if control_latent_channels is None:
            control_latent_channels = self.in_channels

        self.pos_embed_input = comfy.ldm.modules.diffusionmodules.mmdit.PatchEmbed(
            None,
            self.patch_size,
            control_latent_channels,
            self.hidden_size,
            bias=True,
            strict_img_size=False,
            dtype=dtype,
            device=device,
            operations=operations
        )

    def forward(
        self,
        x: torch.Tensor,
        timesteps: torch.Tensor,
        y: Optional[torch.Tensor] = None,
        context: Optional[torch.Tensor] = None,
        hint = None,
    ) -> torch.Tensor:

        #weird sd3 controlnet specific stuff
        y = torch.zeros_like(y)

        if self.context_processor is not None:
            context = self.context_processor(context)

        hw = x.shape[-2:]
        x = self.x_embedder(x) + self.cropped_pos_embed(hw, device=x.device).to(dtype=x.dtype, device=x.device)
        x += self.pos_embed_input(hint)

        c = self.t_embedder(timesteps, dtype=x.dtype)
        if y is not None and self.y_embedder is not None:
            y = self.y_embedder(y)
            c = c + y

        if context is not None:
            context = self.context_embedder(context)

        output = []

        blocks = len(self.joint_blocks)
        for i in range(blocks):
            context, x = self.joint_blocks[i](
                context,
                x,
                c=c,
                use_checkpoint=self.use_checkpoint,
            )

            out = self.controlnet_blocks[i](x)
            count = self.depth // blocks
            if i == blocks - 1:
                count -= 1
            for j in range(count):
                output.append(out)

        return {"output": output}



import math
import os
import sys

import comfy.utils
import cv2
import folder_paths
import model_management
import numpy as np
import torch

# from comfy_extras.chainner_models import model_loading
from hordelib.nodes.facerestore_cf.r_chainner import model_loading
from torchvision.transforms.functional import normalize

from hordelib.nodes.facerestore_cf.basicsr.utils.registry import ARCH_REGISTRY
from hordelib.nodes.facerestore_cf.facelib.detection.retinaface import retinaface
from hordelib.nodes.facerestore_cf.facelib.utils.face_restoration_helper import FaceRestoreHelper

# import codeformer_arch


def img2tensor(imgs, bgr2rgb=True, float32=True):
    """Numpy array to tensor.

    Args:
        imgs (list[ndarray] | ndarray): Input images.
        bgr2rgb (bool): Whether to change bgr to rgb.
        float32 (bool): Whether to change to float32.

    Returns:
        list[tensor] | tensor: Tensor images. If returned results only have
            one element, just return tensor.
    """

    def _totensor(img, bgr2rgb, float32):
        if img.shape[2] == 3 and bgr2rgb:
            if img.dtype == "float64":
                img = img.astype("float32")
            img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
        img = torch.from_numpy(img.transpose(2, 0, 1))
        if float32:
            img = img.float()
        return img

    if isinstance(imgs, list):
        return [_totensor(img, bgr2rgb, float32) for img in imgs]
    else:
        return _totensor(imgs, bgr2rgb, float32)


def tensor2img(tensor, rgb2bgr=True, out_type=np.uint8, min_max=(0, 1)):
    """Convert torch Tensors into image numpy arrays.

    After clamping to [min, max], values will be normalized to [0, 1].

    Args:
        tensor (Tensor or list[Tensor]): Accept shapes:
            1) 4D mini-batch Tensor of shape (B x 3/1 x H x W);
            2) 3D Tensor of shape (3/1 x H x W);
            3) 2D Tensor of shape (H x W).
            Tensor channel should be in RGB order.
        rgb2bgr (bool): Whether to change rgb to bgr.
        out_type (numpy type): output types. If ``np.uint8``, transform outputs
            to uint8 type with range [0, 255]; otherwise, float type with
            range [0, 1]. Default: ``np.uint8``.
        min_max (tuple[int]): min and max values for clamp.

    Returns:
        (Tensor or list): 3D ndarray of shape (H x W x C) OR 2D ndarray of
        shape (H x W). The channel order is BGR.
    """
    if not (torch.is_tensor(tensor) or (isinstance(tensor, list) and all(torch.is_tensor(t) for t in tensor))):
        raise TypeError(f"tensor or list of tensors expected, got {type(tensor)}")

    if torch.is_tensor(tensor):
        tensor = [tensor]
    result = []
    for _tensor in tensor:
        _tensor = _tensor.squeeze(0).float().detach().cpu().clamp_(*min_max)
        _tensor = (_tensor - min_max[0]) / (min_max[1] - min_max[0])

        n_dim = _tensor.dim()
        if n_dim == 4:
            img_np = make_grid(_tensor, nrow=int(math.sqrt(_tensor.size(0))), normalize=False).numpy()
            img_np = img_np.transpose(1, 2, 0)
            if rgb2bgr:
                img_np = cv2.cvtColor(img_np, cv2.COLOR_RGB2BGR)
        elif n_dim == 3:
            img_np = _tensor.numpy()
            img_np = img_np.transpose(1, 2, 0)
            if img_np.shape[2] == 1:  # gray image
                img_np = np.squeeze(img_np, axis=2)
            else:
                if rgb2bgr:
                    img_np = cv2.cvtColor(img_np, cv2.COLOR_RGB2BGR)
        elif n_dim == 2:
            img_np = _tensor.numpy()
        else:
            raise TypeError("Only support 4D, 3D or 2D tensor. " f"But received with dimension: {n_dim}")
        if out_type == np.uint8:
            # Unlike MATLAB, numpy.unit8() WILL NOT round by default.
            img_np = (img_np * 255.0).round()
        img_np = img_np.astype(out_type)
        result.append(img_np)
    if len(result) == 1:
        result = result[0]
    return result


class FaceRestoreCFWithModel:
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "facerestore_model": ("FACERESTORE_MODEL",),
                "image": ("IMAGE",),
                "facedetection": (["retinaface_resnet50", "retinaface_mobile0.25", "YOLOv5l", "YOLOv5n"],),
                "codeformer_fidelity": ("FLOAT", {"default": 0.5, "min": 0.0, "max": 1, "step": 0.05}),
            },
        }

    RETURN_TYPES = ("IMAGE",)

    FUNCTION = "restore_face"

    CATEGORY = "facerestore_cf"

    def __init__(self):
        self.face_helper = None

    def restore_face(self, facerestore_model, image, facedetection, codeformer_fidelity):
        print(f"\tStarting restore_face with codeformer_fidelity: {codeformer_fidelity}")
        device = model_management.get_torch_device()
        facerestore_model.to(device)
        if self.face_helper is None:
            self.face_helper = FaceRestoreHelper(
                1,
                face_size=512,
                crop_ratio=(1, 1),
                det_model=facedetection,
                save_ext="png",
                use_parse=True,
                device=device,
            )

        image_np = 255.0 * image.cpu().numpy()

        total_images = image_np.shape[0]
        out_images = np.ndarray(shape=image_np.shape)

        for i in range(total_images):
            cur_image_np = image_np[i, :, :, ::-1]

            original_resolution = cur_image_np.shape[0:2]

            if facerestore_model is None or self.face_helper is None:
                return image

            self.face_helper.clean_all()
            self.face_helper.read_image(cur_image_np)
            self.face_helper.get_face_landmarks_5(only_center_face=False, resize=640, eye_dist_threshold=5)
            self.face_helper.align_warp_face()

            restored_face = None
            for idx, cropped_face in enumerate(self.face_helper.cropped_faces):
                cropped_face_t = img2tensor(cropped_face / 255.0, bgr2rgb=True, float32=True)
                normalize(cropped_face_t, (0.5, 0.5, 0.5), (0.5, 0.5, 0.5), inplace=True)
                cropped_face_t = cropped_face_t.unsqueeze(0).to(device)

                try:
                    with torch.no_grad():
                        # output = facerestore_model(cropped_face_t, w=strength, adain=True)[0]
                        # output = facerestore_model(cropped_face_t)[0]
                        output = facerestore_model(cropped_face_t, w=codeformer_fidelity)[0]
                        restored_face = tensor2img(output, rgb2bgr=True, min_max=(-1, 1))
                    del output
                    torch.cuda.empty_cache()
                except Exception as error:
                    print(f"\tFailed inference for CodeFormer: {error}", file=sys.stderr)
                    restored_face = tensor2img(cropped_face_t, rgb2bgr=True, min_max=(-1, 1))

                restored_face = restored_face.astype("uint8")
                self.face_helper.add_restored_face(restored_face)

            self.face_helper.get_inverse_affine(None)

            restored_img = self.face_helper.paste_faces_to_input_image()
            restored_img = restored_img[:, :, ::-1]

            if original_resolution != restored_img.shape[0:2]:
                restored_img = cv2.resize(
                    restored_img,
                    (0, 0),
                    fx=original_resolution[1] / restored_img.shape[1],
                    fy=original_resolution[0] / restored_img.shape[0],
                    interpolation=cv2.INTER_LINEAR,
                )

            self.face_helper.clean_all()

            # restored_img = cv2.cvtColor(restored_face, cv2.COLOR_BGR2RGB)

            out_images[i] = restored_img

        restored_img_np = np.array(out_images).astype(np.float32) / 255.0
        restored_img_tensor = torch.from_numpy(restored_img_np)
        return (restored_img_tensor,)


class CropFace:
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "image": ("IMAGE",),
                "facedetection": (["retinaface_resnet50", "retinaface_mobile0.25", "YOLOv5l", "YOLOv5n"],),
            },
        }

    RETURN_TYPES = ("IMAGE",)

    FUNCTION = "crop_face"

    CATEGORY = "facerestore_cf"

    def __init__(self):
        self.face_helper = None

    def crop_face(self, image, facedetection):
        device = model_management.get_torch_device()
        if self.face_helper is None:
            self.face_helper = FaceRestoreHelper(
                1,
                face_size=512,
                crop_ratio=(1, 1),
                det_model=facedetection,
                save_ext="png",
                use_parse=True,
                device=device,
            )

        image_np = 255.0 * image.cpu().numpy()

        total_images = image_np.shape[0]
        out_images = np.ndarray(shape=(total_images, 512, 512, 3))
        next_idx = 0

        for i in range(total_images):

            cur_image_np = image_np[i, :, :, ::-1]

            original_resolution = cur_image_np.shape[0:2]

            if self.face_helper is None:
                return image

            self.face_helper.clean_all()
            self.face_helper.read_image(cur_image_np)
            self.face_helper.get_face_landmarks_5(only_center_face=False, resize=640, eye_dist_threshold=5)
            self.face_helper.align_warp_face()

            faces_found = len(self.face_helper.cropped_faces)
            if faces_found == 0:
                next_idx += 1  # output black image for no face
            if out_images.shape[0] < next_idx + faces_found:
                print(out_images.shape)
                print((next_idx + faces_found, 512, 512, 3))
                print("aaaaa")
                out_images = np.resize(out_images, (next_idx + faces_found, 512, 512, 3))
                print(out_images.shape)
            for j in range(faces_found):
                cropped_face_1 = self.face_helper.cropped_faces[j]
                cropped_face_2 = img2tensor(cropped_face_1 / 255.0, bgr2rgb=True, float32=True)
                normalize(cropped_face_2, (0.5, 0.5, 0.5), (0.5, 0.5, 0.5), inplace=True)
                cropped_face_3 = cropped_face_2.unsqueeze(0).to(device)
                cropped_face_4 = tensor2img(cropped_face_3, rgb2bgr=True, min_max=(-1, 1)).astype("uint8")
                cropped_face_5 = cv2.cvtColor(cropped_face_4, cv2.COLOR_BGR2RGB)
                out_images[next_idx] = cropped_face_5
                next_idx += 1

        cropped_face_6 = np.array(out_images).astype(np.float32) / 255.0
        cropped_face_7 = torch.from_numpy(cropped_face_6)
        return (cropped_face_7,)


class FaceRestoreModelLoader:
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "model_name": (folder_paths.get_filename_list("facerestore_models"),),
            },
        }

    RETURN_TYPES = ("FACERESTORE_MODEL",)
    FUNCTION = "load_model"

    CATEGORY = "facerestore_cf"

    # def load_model(self, model_name):
    #     model_path = folder_paths.get_full_path("facerestore_models", model_name)
    #     sd = comfy.utils.load_torch_file(model_path, safe_load=True)
    #     out = model_loading.load_state_dict(sd).eval()
    #     return (out, )

    def load_model(self, model_name):
        if "codeformer" in model_name.lower():
            print(f"\tLoading CodeFormer: {model_name}")
            model_path = folder_paths.get_full_path("facerestore_models", model_name)
            device = model_management.get_torch_device()
            codeformer_net = ARCH_REGISTRY.get("CodeFormer")(
                dim_embd=512,
                codebook_size=1024,
                n_head=8,
                n_layers=9,
                connect_list=["32", "64", "128", "256"],
            ).to(device)
            checkpoint = torch.load(model_path)["params_ema"]
            codeformer_net.load_state_dict(checkpoint)
            out = codeformer_net.eval()
            return (out,)
        else:
            model_path = folder_paths.get_full_path("facerestore_models", model_name)
            sd = comfy.utils.load_torch_file(model_path, safe_load=True)
            out = model_loading.load_state_dict(sd).eval()
            return (out,)


NODE_CLASS_MAPPINGS = {
    "FaceRestoreCFWithModel": FaceRestoreCFWithModel,
    "CropFace": CropFace,
    "FaceRestoreModelLoader": FaceRestoreModelLoader,
}



#Original code can be found on: https://github.com/XLabs-AI/x-flux/blob/main/src/flux/controlnet.py
#modified to support different types of flux controlnets

import torch
import math
from torch import Tensor, nn
from einops import rearrange, repeat

from .layers import (timestep_embedding)

from .model import Flux
import comfy.ldm.common_dit

class MistolineCondDownsamplBlock(nn.Module):
    def __init__(self, dtype=None, device=None, operations=None):
        super().__init__()
        self.encoder = nn.Sequential(
            operations.Conv2d(3, 16, 3, padding=1, dtype=dtype, device=device),
            nn.SiLU(),
            operations.Conv2d(16, 16, 1, dtype=dtype, device=device),
            nn.SiLU(),
            operations.Conv2d(16, 16, 3, padding=1, dtype=dtype, device=device),
            nn.SiLU(),
            operations.Conv2d(16, 16, 3, padding=1, stride=2, dtype=dtype, device=device),
            nn.SiLU(),
            operations.Conv2d(16, 16, 3, padding=1, dtype=dtype, device=device),
            nn.SiLU(),
            operations.Conv2d(16, 16, 3, padding=1, stride=2, dtype=dtype, device=device),
            nn.SiLU(),
            operations.Conv2d(16, 16, 3, padding=1, dtype=dtype, device=device),
            nn.SiLU(),
            operations.Conv2d(16, 16, 3, padding=1, stride=2, dtype=dtype, device=device),
            nn.SiLU(),
            operations.Conv2d(16, 16, 1, dtype=dtype, device=device),
            nn.SiLU(),
            operations.Conv2d(16, 16, 3, padding=1, dtype=dtype, device=device)
        )

    def forward(self, x):
        return self.encoder(x)

class MistolineControlnetBlock(nn.Module):
    def __init__(self, hidden_size, dtype=None, device=None, operations=None):
        super().__init__()
        self.linear = operations.Linear(hidden_size, hidden_size, dtype=dtype, device=device)
        self.act = nn.SiLU()

    def forward(self, x):
        return self.act(self.linear(x))


class ControlNetFlux(Flux):
    def __init__(self, latent_input=False, num_union_modes=0, mistoline=False, control_latent_channels=None, image_model=None, dtype=None, device=None, operations=None, **kwargs):
        super().__init__(final_layer=False, dtype=dtype, device=device, operations=operations, **kwargs)

        self.main_model_double = 19
        self.main_model_single = 38

        self.mistoline = mistoline
        # add ControlNet blocks
        if self.mistoline:
            control_block = lambda : MistolineControlnetBlock(self.hidden_size, dtype=dtype, device=device, operations=operations)
        else:
            control_block = lambda : operations.Linear(self.hidden_size, self.hidden_size, dtype=dtype, device=device)

        self.controlnet_blocks = nn.ModuleList([])
        for _ in range(self.params.depth):
            self.controlnet_blocks.append(control_block())

        self.controlnet_single_blocks = nn.ModuleList([])
        for _ in range(self.params.depth_single_blocks):
            self.controlnet_single_blocks.append(control_block())

        self.num_union_modes = num_union_modes
        self.controlnet_mode_embedder = None
        if self.num_union_modes > 0:
            self.controlnet_mode_embedder = operations.Embedding(self.num_union_modes, self.hidden_size, dtype=dtype, device=device)

        self.gradient_checkpointing = False
        self.latent_input = latent_input
        if control_latent_channels is None:
            control_latent_channels = self.in_channels
        else:
            control_latent_channels *= 2 * 2 #patch size

        self.pos_embed_input = operations.Linear(control_latent_channels, self.hidden_size, bias=True, dtype=dtype, device=device)
        if not self.latent_input:
            if self.mistoline:
                self.input_cond_block = MistolineCondDownsamplBlock(dtype=dtype, device=device, operations=operations)
            else:
                self.input_hint_block = nn.Sequential(
                    operations.Conv2d(3, 16, 3, padding=1, dtype=dtype, device=device),
                    nn.SiLU(),
                    operations.Conv2d(16, 16, 3, padding=1, dtype=dtype, device=device),
                    nn.SiLU(),
                    operations.Conv2d(16, 16, 3, padding=1, stride=2, dtype=dtype, device=device),
                    nn.SiLU(),
                    operations.Conv2d(16, 16, 3, padding=1, dtype=dtype, device=device),
                    nn.SiLU(),
                    operations.Conv2d(16, 16, 3, padding=1, stride=2, dtype=dtype, device=device),
                    nn.SiLU(),
                    operations.Conv2d(16, 16, 3, padding=1, dtype=dtype, device=device),
                    nn.SiLU(),
                    operations.Conv2d(16, 16, 3, padding=1, stride=2, dtype=dtype, device=device),
                    nn.SiLU(),
                    operations.Conv2d(16, 16, 3, padding=1, dtype=dtype, device=device)
                )

    def forward_orig(
        self,
        img: Tensor,
        img_ids: Tensor,
        controlnet_cond: Tensor,
        txt: Tensor,
        txt_ids: Tensor,
        timesteps: Tensor,
        y: Tensor,
        guidance: Tensor = None,
        control_type: Tensor = None,
    ) -> Tensor:
        if img.ndim != 3 or txt.ndim != 3:
            raise ValueError("Input img and txt tensors must have 3 dimensions.")

        # running on sequences img
        img = self.img_in(img)

        controlnet_cond = self.pos_embed_input(controlnet_cond)
        img = img + controlnet_cond
        vec = self.time_in(timestep_embedding(timesteps, 256))
        if self.params.guidance_embed:
            vec = vec + self.guidance_in(timestep_embedding(guidance, 256))
        vec = vec + self.vector_in(y)
        txt = self.txt_in(txt)

        if self.controlnet_mode_embedder is not None and len(control_type) > 0:
            control_cond = self.controlnet_mode_embedder(torch.tensor(control_type, device=img.device), out_dtype=img.dtype).unsqueeze(0).repeat((txt.shape[0], 1, 1))
            txt = torch.cat([control_cond, txt], dim=1)
            txt_ids = torch.cat([txt_ids[:,:1], txt_ids], dim=1)

        ids = torch.cat((txt_ids, img_ids), dim=1)
        pe = self.pe_embedder(ids)

        controlnet_double = ()

        for i in range(len(self.double_blocks)):
            img, txt = self.double_blocks[i](img=img, txt=txt, vec=vec, pe=pe)
            controlnet_double = controlnet_double + (self.controlnet_blocks[i](img),)

        img = torch.cat((txt, img), 1)

        controlnet_single = ()

        for i in range(len(self.single_blocks)):
            img = self.single_blocks[i](img, vec=vec, pe=pe)
            controlnet_single = controlnet_single + (self.controlnet_single_blocks[i](img[:, txt.shape[1] :, ...]),)

        repeat = math.ceil(self.main_model_double / len(controlnet_double))
        if self.latent_input:
            out_input = ()
            for x in controlnet_double:
                    out_input += (x,) * repeat
        else:
            out_input = (controlnet_double * repeat)

        out = {"input": out_input[:self.main_model_double]}
        if len(controlnet_single) > 0:
            repeat = math.ceil(self.main_model_single / len(controlnet_single))
            out_output = ()
            if self.latent_input:
                for x in controlnet_single:
                        out_output += (x,) * repeat
            else:
                out_output = (controlnet_single * repeat)
            out["output"] = out_output[:self.main_model_single]
        return out

    def forward(self, x, timesteps, context, y, guidance=None, hint=None, **kwargs):
        patch_size = 2
        if self.latent_input:
            hint = comfy.ldm.common_dit.pad_to_patch_size(hint, (patch_size, patch_size))
        elif self.mistoline:
            hint = hint * 2.0 - 1.0
            hint = self.input_cond_block(hint)
        else:
            hint = hint * 2.0 - 1.0
            hint = self.input_hint_block(hint)

        hint = rearrange(hint, "b c (h ph) (w pw) -> b (h w) (c ph pw)", ph=patch_size, pw=patch_size)

        bs, c, h, w = x.shape
        x = comfy.ldm.common_dit.pad_to_patch_size(x, (patch_size, patch_size))

        img = rearrange(x, "b c (h ph) (w pw) -> b (h w) (c ph pw)", ph=patch_size, pw=patch_size)

        h_len = ((h + (patch_size // 2)) // patch_size)
        w_len = ((w + (patch_size // 2)) // patch_size)
        img_ids = torch.zeros((h_len, w_len, 3), device=x.device, dtype=x.dtype)
        img_ids[..., 1] = img_ids[..., 1] + torch.linspace(0, h_len - 1, steps=h_len, device=x.device, dtype=x.dtype)[:, None]
        img_ids[..., 2] = img_ids[..., 2] + torch.linspace(0, w_len - 1, steps=w_len, device=x.device, dtype=x.dtype)[None, :]
        img_ids = repeat(img_ids, "h w c -> b (h w) c", b=bs)

        txt_ids = torch.zeros((bs, context.shape[1], 3), device=x.device, dtype=x.dtype)
        return self.forward_orig(img, img_ids, hint, context, txt_ids, timesteps, y, guidance, control_type=kwargs.get("control_type", []))



import node_helpers

class CLIPTextEncodeFlux:
    @classmethod
    def INPUT_TYPES(s):
        return {"required": {
            "clip": ("CLIP", ),
            "clip_l": ("STRING", {"multiline": True, "dynamicPrompts": True}),
            "t5xxl": ("STRING", {"multiline": True, "dynamicPrompts": True}),
            "guidance": ("FLOAT", {"default": 3.5, "min": 0.0, "max": 100.0, "step": 0.1}),
            }}
    RETURN_TYPES = ("CONDITIONING",)
    FUNCTION = "encode"

    CATEGORY = "advanced/conditioning/flux"

    def encode(self, clip, clip_l, t5xxl, guidance):
        tokens = clip.tokenize(clip_l)
        tokens["t5xxl"] = clip.tokenize(t5xxl)["t5xxl"]

        return (clip.encode_from_tokens_scheduled(tokens, add_dict={"guidance": guidance}), )

class FluxGuidance:
    @classmethod
    def INPUT_TYPES(s):
        return {"required": {
            "conditioning": ("CONDITIONING", ),
            "guidance": ("FLOAT", {"default": 3.5, "min": 0.0, "max": 100.0, "step": 0.1}),
            }}

    RETURN_TYPES = ("CONDITIONING",)
    FUNCTION = "append"

    CATEGORY = "advanced/conditioning/flux"

    def append(self, conditioning, guidance):
        c = node_helpers.conditioning_set_values(conditioning, {"guidance": guidance})
        return (c, )


NODE_CLASS_MAPPINGS = {
    "CLIPTextEncodeFlux": CLIPTextEncodeFlux,
    "FluxGuidance": FluxGuidance,
}



from .deform_conv import (
    DeformConv,
    DeformConvPack,
    ModulatedDeformConv,
    ModulatedDeformConvPack,
    deform_conv,
    modulated_deform_conv,
)

__all__ = [
    "DeformConv",
    "DeformConvPack",
    "ModulatedDeformConv",
    "ModulatedDeformConvPack",
    "deform_conv",
    "modulated_deform_conv",
]



from comfy import sd1_clip
import comfy.text_encoders.sd3_clip
import os
from transformers import T5TokenizerFast


class T5XXLModel(comfy.text_encoders.sd3_clip.T5XXLModel):
    def __init__(self, **kwargs):
        kwargs["attention_mask"] = True
        super().__init__(**kwargs)


class MochiT5XXL(sd1_clip.SD1ClipModel):
    def __init__(self, device="cpu", dtype=None, model_options={}):
        super().__init__(device=device, dtype=dtype, name="t5xxl", clip_model=T5XXLModel, model_options=model_options)


class T5XXLTokenizer(sd1_clip.SDTokenizer):
    def __init__(self, embedding_directory=None, tokenizer_data={}):
        tokenizer_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), "t5_tokenizer")
        super().__init__(tokenizer_path, embedding_directory=embedding_directory, pad_with_end=False, embedding_size=4096, embedding_key='t5xxl', tokenizer_class=T5TokenizerFast, has_start_token=False, pad_to_max_length=False, max_length=99999999, min_length=256)


class MochiT5Tokenizer(sd1_clip.SD1Tokenizer):
    def __init__(self, embedding_directory=None, tokenizer_data={}):
        super().__init__(embedding_directory=embedding_directory, tokenizer_data=tokenizer_data, clip_name="t5xxl", tokenizer=T5XXLTokenizer)


def mochi_te(dtype_t5=None, t5xxl_scaled_fp8=None):
    class MochiTEModel_(MochiT5XXL):
        def __init__(self, device="cpu", dtype=None, model_options={}):
            if t5xxl_scaled_fp8 is not None and "t5xxl_scaled_fp8" not in model_options:
                model_options = model_options.copy()
                model_options["t5xxl_scaled_fp8"] = t5xxl_scaled_fp8
            if dtype is None:
                dtype = dtype_t5
            super().__init__(device=device, dtype=dtype, model_options=model_options)
    return MochiTEModel_



#original code from https://github.com/genmoai/models under apache 2.0 license
#adapted to ComfyUI

from typing import Dict, List, Optional, Tuple

import torch
import torch.nn as nn
import torch.nn.functional as F
from einops import rearrange
# from flash_attn import flash_attn_varlen_qkvpacked_func
from comfy.ldm.modules.attention import optimized_attention

from .layers import (
    FeedForward,
    PatchEmbed,
    RMSNorm,
    TimestepEmbedder,
)

from .rope_mixed import (
    compute_mixed_rotation,
    create_position_matrix,
)
from .temporal_rope import apply_rotary_emb_qk_real
from .utils import (
    AttentionPool,
    modulate,
)

import comfy.ldm.common_dit
import comfy.ops


def modulated_rmsnorm(x, scale, eps=1e-6):
    # Normalize and modulate
    x_normed = comfy.ldm.common_dit.rms_norm(x, eps=eps)
    x_modulated = x_normed * (1 + scale.unsqueeze(1))

    return x_modulated


def residual_tanh_gated_rmsnorm(x, x_res, gate, eps=1e-6):
    # Apply tanh to gate
    tanh_gate = torch.tanh(gate).unsqueeze(1)

    # Normalize and apply gated scaling
    x_normed = comfy.ldm.common_dit.rms_norm(x_res, eps=eps) * tanh_gate

    # Apply residual connection
    output = x + x_normed

    return output

class AsymmetricAttention(nn.Module):
    def __init__(
        self,
        dim_x: int,
        dim_y: int,
        num_heads: int = 8,
        qkv_bias: bool = True,
        qk_norm: bool = False,
        attn_drop: float = 0.0,
        update_y: bool = True,
        out_bias: bool = True,
        attend_to_padding: bool = False,
        softmax_scale: Optional[float] = None,
        device: Optional[torch.device] = None,
        dtype=None,
        operations=None,
    ):
        super().__init__()
        self.dim_x = dim_x
        self.dim_y = dim_y
        self.num_heads = num_heads
        self.head_dim = dim_x // num_heads
        self.attn_drop = attn_drop
        self.update_y = update_y
        self.attend_to_padding = attend_to_padding
        self.softmax_scale = softmax_scale
        if dim_x % num_heads != 0:
            raise ValueError(
                f"dim_x={dim_x} should be divisible by num_heads={num_heads}"
            )

        # Input layers.
        self.qkv_bias = qkv_bias
        self.qkv_x = operations.Linear(dim_x, 3 * dim_x, bias=qkv_bias, device=device, dtype=dtype)
        # Project text features to match visual features (dim_y -> dim_x)
        self.qkv_y = operations.Linear(dim_y, 3 * dim_x, bias=qkv_bias, device=device, dtype=dtype)

        # Query and key normalization for stability.
        assert qk_norm
        self.q_norm_x = RMSNorm(self.head_dim, device=device, dtype=dtype)
        self.k_norm_x = RMSNorm(self.head_dim, device=device, dtype=dtype)
        self.q_norm_y = RMSNorm(self.head_dim, device=device, dtype=dtype)
        self.k_norm_y = RMSNorm(self.head_dim, device=device, dtype=dtype)

        # Output layers. y features go back down from dim_x -> dim_y.
        self.proj_x = operations.Linear(dim_x, dim_x, bias=out_bias, device=device, dtype=dtype)
        self.proj_y = (
            operations.Linear(dim_x, dim_y, bias=out_bias, device=device, dtype=dtype)
            if update_y
            else nn.Identity()
        )

    def forward(
        self,
        x: torch.Tensor,  # (B, N, dim_x)
        y: torch.Tensor,  # (B, L, dim_y)
        scale_x: torch.Tensor,  # (B, dim_x), modulation for pre-RMSNorm.
        scale_y: torch.Tensor,  # (B, dim_y), modulation for pre-RMSNorm.
        crop_y,
        **rope_rotation,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        rope_cos = rope_rotation.get("rope_cos")
        rope_sin = rope_rotation.get("rope_sin")
        # Pre-norm for visual features
        x = modulated_rmsnorm(x, scale_x)  # (B, M, dim_x) where M = N / cp_group_size

        # Process visual features
        # qkv_x = self.qkv_x(x)  # (B, M, 3 * dim_x)
        # assert qkv_x.dtype == torch.bfloat16
        # qkv_x = all_to_all_collect_tokens(
        #     qkv_x, self.num_heads
        # )  # (3, B, N, local_h, head_dim)

        # Process text features
        y = modulated_rmsnorm(y, scale_y)  # (B, L, dim_y)
        q_y, k_y, v_y = self.qkv_y(y).view(y.shape[0], y.shape[1], 3, self.num_heads, -1).unbind(2)  # (B, N, local_h, head_dim)

        q_y = self.q_norm_y(q_y)
        k_y = self.k_norm_y(k_y)

        # Split qkv_x into q, k, v
        q_x, k_x, v_x = self.qkv_x(x).view(x.shape[0], x.shape[1], 3, self.num_heads, -1).unbind(2)  # (B, N, local_h, head_dim)
        q_x = self.q_norm_x(q_x)
        q_x = apply_rotary_emb_qk_real(q_x, rope_cos, rope_sin)
        k_x = self.k_norm_x(k_x)
        k_x = apply_rotary_emb_qk_real(k_x, rope_cos, rope_sin)

        q = torch.cat([q_x, q_y[:, :crop_y]], dim=1).transpose(1, 2)
        k = torch.cat([k_x, k_y[:, :crop_y]], dim=1).transpose(1, 2)
        v = torch.cat([v_x, v_y[:, :crop_y]], dim=1).transpose(1, 2)

        xy = optimized_attention(q,
                                 k,
                                 v, self.num_heads, skip_reshape=True)

        x, y = torch.tensor_split(xy, (q_x.shape[1],), dim=1)
        x = self.proj_x(x)
        o = torch.zeros(y.shape[0], q_y.shape[1], y.shape[-1], device=y.device, dtype=y.dtype)
        o[:, :y.shape[1]] = y

        y = self.proj_y(o)
        # print("ox", x)
        # print("oy", y)
        return x, y


class AsymmetricJointBlock(nn.Module):
    def __init__(
        self,
        hidden_size_x: int,
        hidden_size_y: int,
        num_heads: int,
        *,
        mlp_ratio_x: float = 8.0,  # Ratio of hidden size to d_model for MLP for visual tokens.
        mlp_ratio_y: float = 4.0,  # Ratio of hidden size to d_model for MLP for text tokens.
        update_y: bool = True,  # Whether to update text tokens in this block.
        device: Optional[torch.device] = None,
        dtype=None,
        operations=None,
        **block_kwargs,
    ):
        super().__init__()
        self.update_y = update_y
        self.hidden_size_x = hidden_size_x
        self.hidden_size_y = hidden_size_y
        self.mod_x = operations.Linear(hidden_size_x, 4 * hidden_size_x, device=device, dtype=dtype)
        if self.update_y:
            self.mod_y = operations.Linear(hidden_size_x, 4 * hidden_size_y, device=device, dtype=dtype)
        else:
            self.mod_y = operations.Linear(hidden_size_x, hidden_size_y, device=device, dtype=dtype)

        # Self-attention:
        self.attn = AsymmetricAttention(
            hidden_size_x,
            hidden_size_y,
            num_heads=num_heads,
            update_y=update_y,
            device=device,
            dtype=dtype,
            operations=operations,
            **block_kwargs,
        )

        # MLP.
        mlp_hidden_dim_x = int(hidden_size_x * mlp_ratio_x)
        assert mlp_hidden_dim_x == int(1536 * 8)
        self.mlp_x = FeedForward(
            in_features=hidden_size_x,
            hidden_size=mlp_hidden_dim_x,
            multiple_of=256,
            ffn_dim_multiplier=None,
            device=device,
            dtype=dtype,
            operations=operations,
        )

        # MLP for text not needed in last block.
        if self.update_y:
            mlp_hidden_dim_y = int(hidden_size_y * mlp_ratio_y)
            self.mlp_y = FeedForward(
                in_features=hidden_size_y,
                hidden_size=mlp_hidden_dim_y,
                multiple_of=256,
                ffn_dim_multiplier=None,
                device=device,
                dtype=dtype,
                operations=operations,
            )

    def forward(
        self,
        x: torch.Tensor,
        c: torch.Tensor,
        y: torch.Tensor,
        **attn_kwargs,
    ):
        """Forward pass of a block.

        Args:
            x: (B, N, dim) tensor of visual tokens
            c: (B, dim) tensor of conditioned features
            y: (B, L, dim) tensor of text tokens
            num_frames: Number of frames in the video. N = num_frames * num_spatial_tokens

        Returns:
            x: (B, N, dim) tensor of visual tokens after block
            y: (B, L, dim) tensor of text tokens after block
        """
        N = x.size(1)

        c = F.silu(c)
        mod_x = self.mod_x(c)
        scale_msa_x, gate_msa_x, scale_mlp_x, gate_mlp_x = mod_x.chunk(4, dim=1)

        mod_y = self.mod_y(c)
        if self.update_y:
            scale_msa_y, gate_msa_y, scale_mlp_y, gate_mlp_y = mod_y.chunk(4, dim=1)
        else:
            scale_msa_y = mod_y

        # Self-attention block.
        x_attn, y_attn = self.attn(
            x,
            y,
            scale_x=scale_msa_x,
            scale_y=scale_msa_y,
            **attn_kwargs,
        )

        assert x_attn.size(1) == N
        x = residual_tanh_gated_rmsnorm(x, x_attn, gate_msa_x)
        if self.update_y:
            y = residual_tanh_gated_rmsnorm(y, y_attn, gate_msa_y)

        # MLP block.
        x = self.ff_block_x(x, scale_mlp_x, gate_mlp_x)
        if self.update_y:
            y = self.ff_block_y(y, scale_mlp_y, gate_mlp_y)

        return x, y

    def ff_block_x(self, x, scale_x, gate_x):
        x_mod = modulated_rmsnorm(x, scale_x)
        x_res = self.mlp_x(x_mod)
        x = residual_tanh_gated_rmsnorm(x, x_res, gate_x)  # Sandwich norm
        return x

    def ff_block_y(self, y, scale_y, gate_y):
        y_mod = modulated_rmsnorm(y, scale_y)
        y_res = self.mlp_y(y_mod)
        y = residual_tanh_gated_rmsnorm(y, y_res, gate_y)  # Sandwich norm
        return y


class FinalLayer(nn.Module):
    """
    The final layer of DiT.
    """

    def __init__(
        self,
        hidden_size,
        patch_size,
        out_channels,
        device: Optional[torch.device] = None,
        dtype=None,
        operations=None,
    ):
        super().__init__()
        self.norm_final = operations.LayerNorm(
            hidden_size, elementwise_affine=False, eps=1e-6, device=device, dtype=dtype
        )
        self.mod = operations.Linear(hidden_size, 2 * hidden_size, device=device, dtype=dtype)
        self.linear = operations.Linear(
            hidden_size, patch_size * patch_size * out_channels, device=device, dtype=dtype
        )

    def forward(self, x, c):
        c = F.silu(c)
        shift, scale = self.mod(c).chunk(2, dim=1)
        x = modulate(self.norm_final(x), shift, scale)
        x = self.linear(x)
        return x


class AsymmDiTJoint(nn.Module):
    """
    Diffusion model with a Transformer backbone.

    Ingests text embeddings instead of a label.
    """

    def __init__(
        self,
        *,
        patch_size=2,
        in_channels=4,
        hidden_size_x=1152,
        hidden_size_y=1152,
        depth=48,
        num_heads=16,
        mlp_ratio_x=8.0,
        mlp_ratio_y=4.0,
        use_t5: bool = False,
        t5_feat_dim: int = 4096,
        t5_token_length: int = 256,
        learn_sigma=True,
        patch_embed_bias: bool = True,
        timestep_mlp_bias: bool = True,
        attend_to_padding: bool = False,
        timestep_scale: Optional[float] = None,
        use_extended_posenc: bool = False,
        posenc_preserve_area: bool = False,
        rope_theta: float = 10000.0,
        image_model=None,
        device: Optional[torch.device] = None,
        dtype=None,
        operations=None,
        **block_kwargs,
    ):
        super().__init__()

        self.dtype = dtype
        self.learn_sigma = learn_sigma
        self.in_channels = in_channels
        self.out_channels = in_channels * 2 if learn_sigma else in_channels
        self.patch_size = patch_size
        self.num_heads = num_heads
        self.hidden_size_x = hidden_size_x
        self.hidden_size_y = hidden_size_y
        self.head_dim = (
            hidden_size_x // num_heads
        )  # Head dimension and count is determined by visual.
        self.attend_to_padding = attend_to_padding
        self.use_extended_posenc = use_extended_posenc
        self.posenc_preserve_area = posenc_preserve_area
        self.use_t5 = use_t5
        self.t5_token_length = t5_token_length
        self.t5_feat_dim = t5_feat_dim
        self.rope_theta = (
            rope_theta  # Scaling factor for frequency computation for temporal RoPE.
        )

        self.x_embedder = PatchEmbed(
            patch_size=patch_size,
            in_chans=in_channels,
            embed_dim=hidden_size_x,
            bias=patch_embed_bias,
            dtype=dtype,
            device=device,
            operations=operations
        )
        # Conditionings
        # Timestep
        self.t_embedder = TimestepEmbedder(
            hidden_size_x, bias=timestep_mlp_bias, timestep_scale=timestep_scale, dtype=dtype, device=device, operations=operations
        )

        if self.use_t5:
            # Caption Pooling (T5)
            self.t5_y_embedder = AttentionPool(
                t5_feat_dim, num_heads=8, output_dim=hidden_size_x, dtype=dtype, device=device, operations=operations
            )

            # Dense Embedding Projection (T5)
            self.t5_yproj = operations.Linear(
                t5_feat_dim, hidden_size_y, bias=True, dtype=dtype, device=device
            )

        # Initialize pos_frequencies as an empty parameter.
        self.pos_frequencies = nn.Parameter(
            torch.empty(3, self.num_heads, self.head_dim // 2, dtype=dtype, device=device)
        )

        assert not self.attend_to_padding

        # for depth 48:
        #  b =  0: AsymmetricJointBlock, update_y=True
        #  b =  1: AsymmetricJointBlock, update_y=True
        #  ...
        #  b = 46: AsymmetricJointBlock, update_y=True
        #  b = 47: AsymmetricJointBlock, update_y=False. No need to update text features.
        blocks = []
        for b in range(depth):
            # Joint multi-modal block
            update_y = b < depth - 1
            block = AsymmetricJointBlock(
                hidden_size_x,
                hidden_size_y,
                num_heads,
                mlp_ratio_x=mlp_ratio_x,
                mlp_ratio_y=mlp_ratio_y,
                update_y=update_y,
                attend_to_padding=attend_to_padding,
                device=device,
                dtype=dtype,
                operations=operations,
                **block_kwargs,
            )

            blocks.append(block)
        self.blocks = nn.ModuleList(blocks)

        self.final_layer = FinalLayer(
            hidden_size_x, patch_size, self.out_channels, dtype=dtype, device=device, operations=operations
        )

    def embed_x(self, x: torch.Tensor) -> torch.Tensor:
        """
        Args:
            x: (B, C=12, T, H, W) tensor of visual tokens

        Returns:
            x: (B, C=3072, N) tensor of visual tokens with positional embedding.
        """
        return self.x_embedder(x)  # Convert BcTHW to BCN

    def prepare(
        self,
        x: torch.Tensor,
        sigma: torch.Tensor,
        t5_feat: torch.Tensor,
        t5_mask: torch.Tensor,
    ):
        """Prepare input and conditioning embeddings."""
        # Visual patch embeddings with positional encoding.
        T, H, W = x.shape[-3:]
        pH, pW = H // self.patch_size, W // self.patch_size
        x = self.embed_x(x)  # (B, N, D), where N = T * H * W / patch_size ** 2
        assert x.ndim == 3

        pH, pW = H // self.patch_size, W // self.patch_size
        N = T * pH * pW
        assert x.size(1) == N
        pos = create_position_matrix(
            T, pH=pH, pW=pW, device=x.device, dtype=torch.float32
        )  # (N, 3)
        rope_cos, rope_sin = compute_mixed_rotation(
            freqs=comfy.ops.cast_to(self.pos_frequencies, dtype=x.dtype, device=x.device), pos=pos
        )  # Each are (N, num_heads, dim // 2)

        c_t = self.t_embedder(1 - sigma, out_dtype=x.dtype)  # (B, D)

        t5_y_pool = self.t5_y_embedder(t5_feat, t5_mask)  # (B, D)

        c = c_t + t5_y_pool

        y_feat = self.t5_yproj(t5_feat)  # (B, L, t5_feat_dim) --> (B, L, D)

        return x, c, y_feat, rope_cos, rope_sin

    def forward(
        self,
        x: torch.Tensor,
        timestep: torch.Tensor,
        context: List[torch.Tensor],
        attention_mask: List[torch.Tensor],
        num_tokens=256,
        packed_indices: Dict[str, torch.Tensor] = None,
        rope_cos: torch.Tensor = None,
        rope_sin: torch.Tensor = None,
        control=None, transformer_options={}, **kwargs
    ):
        patches_replace = transformer_options.get("patches_replace", {})
        y_feat = context
        y_mask = attention_mask
        sigma = timestep
        """Forward pass of DiT.

        Args:
            x: (B, C, T, H, W) tensor of spatial inputs (images or latent representations of images)
            sigma: (B,) tensor of noise standard deviations
            y_feat: List((B, L, y_feat_dim) tensor of caption token features. For SDXL text encoders: L=77, y_feat_dim=2048)
            y_mask: List((B, L) boolean tensor indicating which tokens are not padding)
            packed_indices: Dict with keys for Flash Attention. Result of compute_packed_indices.
        """
        B, _, T, H, W = x.shape

        x, c, y_feat, rope_cos, rope_sin = self.prepare(
            x, sigma, y_feat, y_mask
        )
        del y_mask

        blocks_replace = patches_replace.get("dit", {})
        for i, block in enumerate(self.blocks):
            if ("double_block", i) in blocks_replace:
                def block_wrap(args):
                    out = {}
                    out["img"], out["txt"] = block(
                                                    args["img"],
                                                    args["vec"],
                                                    args["txt"],
                                                    rope_cos=args["rope_cos"],
                                                    rope_sin=args["rope_sin"],
                                                    crop_y=args["num_tokens"]
                                                    )
                    return out
                out = blocks_replace[("double_block", i)]({"img": x, "txt": y_feat, "vec": c, "rope_cos": rope_cos, "rope_sin": rope_sin, "num_tokens": num_tokens}, {"original_block": block_wrap})
                y_feat = out["txt"]
                x = out["img"]
            else:
                x, y_feat = block(
                    x,
                    c,
                    y_feat,
                    rope_cos=rope_cos,
                    rope_sin=rope_sin,
                    crop_y=num_tokens,
                )  # (B, M, D), (B, L, D)
        del y_feat  # Final layers don't use dense text features.

        x = self.final_layer(x, c)  # (B, M, patch_size ** 2 * out_channels)
        x = rearrange(
            x,
            "B (T hp wp) (p1 p2 c) -> B c T (hp p1) (wp p2)",
            T=T,
            hp=H // self.patch_size,
            wp=W // self.patch_size,
            p1=self.patch_size,
            p2=self.patch_size,
            c=self.out_channels,
        )

        return -x



import logging
from spandrel import ModelLoader

def load_state_dict(state_dict):
    logging.warning("comfy_extras.chainner_models is deprecated and has been replaced by the spandrel library.")
    return ModelLoader().load_from_state_dict(state_dict).eval()



from abc import ABC, abstractmethod
from typing import Tuple

import torch
from einops import rearrange
from torch import Tensor


def append_dims(x: torch.Tensor, target_dims: int) -> torch.Tensor:
    """Appends dimensions to the end of a tensor until it has target_dims dimensions."""
    dims_to_append = target_dims - x.ndim
    if dims_to_append < 0:
        raise ValueError(
            f"input has {x.ndim} dims but target_dims is {target_dims}, which is less"
        )
    elif dims_to_append == 0:
        return x
    return x[(...,) + (None,) * dims_to_append]


class Patchifier(ABC):
    def __init__(self, patch_size: int):
        super().__init__()
        self._patch_size = (1, patch_size, patch_size)

    @abstractmethod
    def patchify(
        self, latents: Tensor, frame_rates: Tensor, scale_grid: bool
    ) -> Tuple[Tensor, Tensor]:
        pass

    @abstractmethod
    def unpatchify(
        self,
        latents: Tensor,
        output_height: int,
        output_width: int,
        output_num_frames: int,
        out_channels: int,
    ) -> Tuple[Tensor, Tensor]:
        pass

    @property
    def patch_size(self):
        return self._patch_size

    def get_grid(
        self, orig_num_frames, orig_height, orig_width, batch_size, scale_grid, device
    ):
        f = orig_num_frames // self._patch_size[0]
        h = orig_height // self._patch_size[1]
        w = orig_width // self._patch_size[2]
        grid_h = torch.arange(h, dtype=torch.float32, device=device)
        grid_w = torch.arange(w, dtype=torch.float32, device=device)
        grid_f = torch.arange(f, dtype=torch.float32, device=device)
        grid = torch.meshgrid(grid_f, grid_h, grid_w, indexing='ij')
        grid = torch.stack(grid, dim=0)
        grid = grid.unsqueeze(0).repeat(batch_size, 1, 1, 1, 1)

        if scale_grid is not None:
            for i in range(3):
                if isinstance(scale_grid[i], Tensor):
                    scale = append_dims(scale_grid[i], grid.ndim - 1)
                else:
                    scale = scale_grid[i]
                grid[:, i, ...] = grid[:, i, ...] * scale * self._patch_size[i]

        grid = rearrange(grid, "b c f h w -> b c (f h w)", b=batch_size)
        return grid


class SymmetricPatchifier(Patchifier):
    def patchify(
        self,
        latents: Tensor,
    ) -> Tuple[Tensor, Tensor]:
        latents = rearrange(
            latents,
            "b c (f p1) (h p2) (w p3) -> b (f h w) (c p1 p2 p3)",
            p1=self._patch_size[0],
            p2=self._patch_size[1],
            p3=self._patch_size[2],
        )
        return latents

    def unpatchify(
        self,
        latents: Tensor,
        output_height: int,
        output_width: int,
        output_num_frames: int,
        out_channels: int,
    ) -> Tuple[Tensor, Tensor]:
        output_height = output_height // self._patch_size[1]
        output_width = output_width // self._patch_size[2]
        latents = rearrange(
            latents,
            "b (f h w) (c p q) -> b c f (h p) (w q) ",
            f=output_num_frames,
            h=output_height,
            w=output_width,
            p=self._patch_size[1],
            q=self._patch_size[2],
        )
        return latents



from .specific_tests import TEST_NODE_CLASS_MAPPINGS, TEST_NODE_DISPLAY_NAME_MAPPINGS
from .flow_control import FLOW_CONTROL_NODE_CLASS_MAPPINGS, FLOW_CONTROL_NODE_DISPLAY_NAME_MAPPINGS
from .util import UTILITY_NODE_CLASS_MAPPINGS, UTILITY_NODE_DISPLAY_NAME_MAPPINGS
from .conditions import CONDITION_NODE_CLASS_MAPPINGS, CONDITION_NODE_DISPLAY_NAME_MAPPINGS
from .stubs import TEST_STUB_NODE_CLASS_MAPPINGS, TEST_STUB_NODE_DISPLAY_NAME_MAPPINGS

# NODE_CLASS_MAPPINGS = GENERAL_NODE_CLASS_MAPPINGS.update(COMPONENT_NODE_CLASS_MAPPINGS)
# NODE_DISPLAY_NAME_MAPPINGS = GENERAL_NODE_DISPLAY_NAME_MAPPINGS.update(COMPONENT_NODE_DISPLAY_NAME_MAPPINGS)

NODE_CLASS_MAPPINGS = {}
NODE_CLASS_MAPPINGS.update(TEST_NODE_CLASS_MAPPINGS)
NODE_CLASS_MAPPINGS.update(FLOW_CONTROL_NODE_CLASS_MAPPINGS)
NODE_CLASS_MAPPINGS.update(UTILITY_NODE_CLASS_MAPPINGS)
NODE_CLASS_MAPPINGS.update(CONDITION_NODE_CLASS_MAPPINGS)
NODE_CLASS_MAPPINGS.update(TEST_STUB_NODE_CLASS_MAPPINGS)

NODE_DISPLAY_NAME_MAPPINGS = {}
NODE_DISPLAY_NAME_MAPPINGS.update(TEST_NODE_DISPLAY_NAME_MAPPINGS)
NODE_DISPLAY_NAME_MAPPINGS.update(FLOW_CONTROL_NODE_DISPLAY_NAME_MAPPINGS)
NODE_DISPLAY_NAME_MAPPINGS.update(UTILITY_NODE_DISPLAY_NAME_MAPPINGS)
NODE_DISPLAY_NAME_MAPPINGS.update(CONDITION_NODE_DISPLAY_NAME_MAPPINGS)
NODE_DISPLAY_NAME_MAPPINGS.update(TEST_STUB_NODE_DISPLAY_NAME_MAPPINGS)




from typing import Dict, List, Optional
from api_server.utils.file_operations import FileSystemOperations, FileSystemItem

class FileService:
    def __init__(self, allowed_directories: Dict[str, str], file_system_ops: Optional[FileSystemOperations] = None):
        self.allowed_directories: Dict[str, str] = allowed_directories
        self.file_system_ops: FileSystemOperations = file_system_ops or FileSystemOperations()

    def list_files(self, directory_key: str) -> List[FileSystemItem]:
        if directory_key not in self.allowed_directories:
            raise ValueError("Invalid directory key")
        directory_path: str = self.allowed_directories[directory_key]
        return self.file_system_ops.walk_directory(directory_path)


from comfy import sd1_clip
import os

class LongClipTokenizer_(sd1_clip.SDTokenizer):
    def __init__(self, embedding_directory=None, tokenizer_data={}):
        super().__init__(max_length=248, embedding_directory=embedding_directory, tokenizer_data=tokenizer_data)

class LongClipModel_(sd1_clip.SDClipModel):
    def __init__(self, *args, **kwargs):
        textmodel_json_config = os.path.join(os.path.dirname(os.path.realpath(__file__)), "long_clipl.json")
        super().__init__(*args, textmodel_json_config=textmodel_json_config, **kwargs)

class LongClipTokenizer(sd1_clip.SD1Tokenizer):
    def __init__(self, embedding_directory=None, tokenizer_data={}):
        super().__init__(embedding_directory=embedding_directory, tokenizer_data=tokenizer_data, tokenizer=LongClipTokenizer_)

class LongClipModel(sd1_clip.SD1ClipModel):
    def __init__(self, device="cpu", dtype=None, model_options={}, **kwargs):
        super().__init__(device=device, dtype=dtype, model_options=model_options, clip_model=LongClipModel_, **kwargs)

def model_options_long_clip(sd, tokenizer_data, model_options):
    w = sd.get("clip_l.text_model.embeddings.position_embedding.weight", None)
    if w is None:
        w = sd.get("text_model.embeddings.position_embedding.weight", None)
    if w is not None and w.shape[0] == 248:
        tokenizer_data = tokenizer_data.copy()
        model_options = model_options.copy()
        tokenizer_data["clip_l_tokenizer_class"] = LongClipTokenizer_
        model_options["clip_l_class"] = LongClipModel_
    return tokenizer_data, model_options



import torch
from .tools import VariantSupport
from comfy_execution.graph_utils import GraphBuilder

class TestLazyMixImages:
    @classmethod
    def INPUT_TYPES(cls):
        return {
            "required": {
                "image1": ("IMAGE",{"lazy": True}),
                "image2": ("IMAGE",{"lazy": True}),
                "mask": ("MASK",),
            },
        }

    RETURN_TYPES = ("IMAGE",)
    FUNCTION = "mix"

    CATEGORY = "Testing/Nodes"

    def check_lazy_status(self, mask, image1, image2):
        mask_min = mask.min()
        mask_max = mask.max()
        needed = []
        if image1 is None and (mask_min != 1.0 or mask_max != 1.0):
            needed.append("image1")
        if image2 is None and (mask_min != 0.0 or mask_max != 0.0):
            needed.append("image2")
        return needed

    # Not trying to handle different batch sizes here just to keep the demo simple
    def mix(self, mask, image1, image2):
        mask_min = mask.min()
        mask_max = mask.max()
        if mask_min == 0.0 and mask_max == 0.0:
            return (image1,)
        elif mask_min == 1.0 and mask_max == 1.0:
            return (image2,)

        if len(mask.shape) == 2:
            mask = mask.unsqueeze(0)
        if len(mask.shape) == 3:
            mask = mask.unsqueeze(3)
        if mask.shape[3] < image1.shape[3]:
            mask = mask.repeat(1, 1, 1, image1.shape[3])

        result = image1 * (1. - mask) + image2 * mask,
        return (result[0],)

class TestVariadicAverage:
    @classmethod
    def INPUT_TYPES(cls):
        return {
            "required": {
                "input1": ("IMAGE",),
            },
        }

    RETURN_TYPES = ("IMAGE",)
    FUNCTION = "variadic_average"

    CATEGORY = "Testing/Nodes"

    def variadic_average(self, input1, **kwargs):
        inputs = [input1]
        while 'input' + str(len(inputs) + 1) in kwargs:
            inputs.append(kwargs['input' + str(len(inputs) + 1)])
        return (torch.stack(inputs).mean(dim=0),)


class TestCustomIsChanged:
    @classmethod
    def INPUT_TYPES(cls):
        return {
            "required": {
                "image": ("IMAGE",),
            },
            "optional": {
                "should_change": ("BOOL", {"default": False}),
            },
        }

    RETURN_TYPES = ("IMAGE",)
    FUNCTION = "custom_is_changed"

    CATEGORY = "Testing/Nodes"

    def custom_is_changed(self, image, should_change=False):
        return (image,)
    
    @classmethod
    def IS_CHANGED(cls, should_change=False, *args, **kwargs):
        if should_change:
            return float("NaN")
        else:
            return False

class TestIsChangedWithConstants:
    @classmethod
    def INPUT_TYPES(cls):
        return {
            "required": {
                "image": ("IMAGE",),
                "value": ("FLOAT", {"default": 1.0, "min": 0.0, "max": 10.0}),
            },
        }

    RETURN_TYPES = ("IMAGE",)
    FUNCTION = "custom_is_changed"

    CATEGORY = "Testing/Nodes"

    def custom_is_changed(self, image, value):
        return (image * value,)
    
    @classmethod
    def IS_CHANGED(cls, image, value):
        if image is None:
            return value
        else:
            return image.mean().item() * value

class TestCustomValidation1:
    @classmethod
    def INPUT_TYPES(cls):
        return {
            "required": {
                "input1": ("IMAGE,FLOAT",),
                "input2": ("IMAGE,FLOAT",),
            },
        }

    RETURN_TYPES = ("IMAGE",)
    FUNCTION = "custom_validation1"

    CATEGORY = "Testing/Nodes"

    def custom_validation1(self, input1, input2):
        if isinstance(input1, float) and isinstance(input2, float):
            result = torch.ones([1, 512, 512, 3]) * input1 * input2
        else:
            result = input1 * input2
        return (result,)

    @classmethod
    def VALIDATE_INPUTS(cls, input1=None, input2=None):
        if input1 is not None:
            if not isinstance(input1, (torch.Tensor, float)):
                return f"Invalid type of input1: {type(input1)}"
        if input2 is not None:
            if not isinstance(input2, (torch.Tensor, float)):
                return f"Invalid type of input2: {type(input2)}"

        return True

class TestCustomValidation2:
    @classmethod
    def INPUT_TYPES(cls):
        return {
            "required": {
                "input1": ("IMAGE,FLOAT",),
                "input2": ("IMAGE,FLOAT",),
            },
        }

    RETURN_TYPES = ("IMAGE",)
    FUNCTION = "custom_validation2"

    CATEGORY = "Testing/Nodes"

    def custom_validation2(self, input1, input2):
        if isinstance(input1, float) and isinstance(input2, float):
            result = torch.ones([1, 512, 512, 3]) * input1 * input2
        else:
            result = input1 * input2
        return (result,)

    @classmethod
    def VALIDATE_INPUTS(cls, input_types, input1=None, input2=None):
        if input1 is not None:
            if not isinstance(input1, (torch.Tensor, float)):
                return f"Invalid type of input1: {type(input1)}"
        if input2 is not None:
            if not isinstance(input2, (torch.Tensor, float)):
                return f"Invalid type of input2: {type(input2)}"

        if 'input1' in input_types:
            if input_types['input1'] not in ["IMAGE", "FLOAT"]:
                return f"Invalid type of input1: {input_types['input1']}"
        if 'input2' in input_types:
            if input_types['input2'] not in ["IMAGE", "FLOAT"]:
                return f"Invalid type of input2: {input_types['input2']}"

        return True

@VariantSupport()
class TestCustomValidation3:
    @classmethod
    def INPUT_TYPES(cls):
        return {
            "required": {
                "input1": ("IMAGE,FLOAT",),
                "input2": ("IMAGE,FLOAT",),
            },
        }

    RETURN_TYPES = ("IMAGE",)
    FUNCTION = "custom_validation3"

    CATEGORY = "Testing/Nodes"

    def custom_validation3(self, input1, input2):
        if isinstance(input1, float) and isinstance(input2, float):
            result = torch.ones([1, 512, 512, 3]) * input1 * input2
        else:
            result = input1 * input2
        return (result,)

class TestCustomValidation4:
    @classmethod
    def INPUT_TYPES(cls):
        return {
            "required": {
                "input1": ("FLOAT",),
                "input2": ("FLOAT",),
            },
        }

    RETURN_TYPES = ("IMAGE",)
    FUNCTION = "custom_validation4"

    CATEGORY = "Testing/Nodes"

    def custom_validation4(self, input1, input2):
        result = torch.ones([1, 512, 512, 3]) * input1 * input2
        return (result,)

    @classmethod
    def VALIDATE_INPUTS(cls, input1, input2):
        if input1 is not None:
            if not isinstance(input1, float):
                return f"Invalid type of input1: {type(input1)}"
        if input2 is not None:
            if not isinstance(input2, float):
                return f"Invalid type of input2: {type(input2)}"

        return True

class TestCustomValidation5:
    @classmethod
    def INPUT_TYPES(cls):
        return {
            "required": {
                "input1": ("FLOAT", {"min": 0.0, "max": 1.0}),
                "input2": ("FLOAT", {"min": 0.0, "max": 1.0}),
            },
        }

    RETURN_TYPES = ("IMAGE",)
    FUNCTION = "custom_validation5"

    CATEGORY = "Testing/Nodes"

    def custom_validation5(self, input1, input2):
        value = input1 * input2
        return (torch.ones([1, 512, 512, 3]) * value,)

    @classmethod
    def VALIDATE_INPUTS(cls, **kwargs):
        if kwargs['input2'] == 7.0:
            return "7s are not allowed. I've never liked 7s."
        return True

class TestDynamicDependencyCycle:
    @classmethod
    def INPUT_TYPES(cls):
        return {
            "required": {
                "input1": ("IMAGE",),
                "input2": ("IMAGE",),
            },
        }

    RETURN_TYPES = ("IMAGE",)
    FUNCTION = "dynamic_dependency_cycle"

    CATEGORY = "Testing/Nodes"

    def dynamic_dependency_cycle(self, input1, input2):
        g = GraphBuilder()
        mask = g.node("StubMask", value=0.5, height=512, width=512, batch_size=1)
        mix1 = g.node("TestLazyMixImages", image1=input1, mask=mask.out(0))
        mix2 = g.node("TestLazyMixImages", image1=mix1.out(0), image2=input2, mask=mask.out(0))

        # Create the cyle
        mix1.set_input("image2", mix2.out(0))

        return {
            "result": (mix2.out(0),),
            "expand": g.finalize(),
        }

class TestMixedExpansionReturns:
    @classmethod
    def INPUT_TYPES(cls):
        return {
            "required": {
                "input1": ("FLOAT",),
            },
        }

    RETURN_TYPES = ("IMAGE","IMAGE")
    FUNCTION = "mixed_expansion_returns"

    CATEGORY = "Testing/Nodes"

    def mixed_expansion_returns(self, input1):
        white_image = torch.ones([1, 512, 512, 3])
        if input1 <= 0.1:
            return (torch.ones([1, 512, 512, 3]) * 0.1, white_image)
        elif input1 <= 0.2:
            return {
                "result": (torch.ones([1, 512, 512, 3]) * 0.2, white_image),
            }
        else:
            g = GraphBuilder()
            mask = g.node("StubMask", value=0.3, height=512, width=512, batch_size=1)
            black = g.node("StubImage", content="BLACK", height=512, width=512, batch_size=1)
            white = g.node("StubImage", content="WHITE", height=512, width=512, batch_size=1)
            mix = g.node("TestLazyMixImages", image1=black.out(0), image2=white.out(0), mask=mask.out(0))
            return {
                "result": (mix.out(0), white_image),
                "expand": g.finalize(),
            }

TEST_NODE_CLASS_MAPPINGS = {
    "TestLazyMixImages": TestLazyMixImages,
    "TestVariadicAverage": TestVariadicAverage,
    "TestCustomIsChanged": TestCustomIsChanged,
    "TestIsChangedWithConstants": TestIsChangedWithConstants,
    "TestCustomValidation1": TestCustomValidation1,
    "TestCustomValidation2": TestCustomValidation2,
    "TestCustomValidation3": TestCustomValidation3,
    "TestCustomValidation4": TestCustomValidation4,
    "TestCustomValidation5": TestCustomValidation5,
    "TestDynamicDependencyCycle": TestDynamicDependencyCycle,
    "TestMixedExpansionReturns": TestMixedExpansionReturns,
}

TEST_NODE_DISPLAY_NAME_MAPPINGS = {
    "TestLazyMixImages": "Lazy Mix Images",
    "TestVariadicAverage": "Variadic Average",
    "TestCustomIsChanged": "Custom IsChanged",
    "TestIsChangedWithConstants": "IsChanged With Constants",
    "TestCustomValidation1": "Custom Validation 1",
    "TestCustomValidation2": "Custom Validation 2",
    "TestCustomValidation3": "Custom Validation 3",
    "TestCustomValidation4": "Custom Validation 4",
    "TestCustomValidation5": "Custom Validation 5",
    "TestDynamicDependencyCycle": "Dynamic Dependency Cycle",
    "TestMixedExpansionReturns": "Mixed Expansion Returns",
}
